#!/usr/bin/python

import sys
import system
import behaviour
import fdr
import os
import predicates
import systemeditor
from logwindow import LogWindow
import gtk
import processexplorer
import util
import errno
import refinementchecker
import gnome

WINDOW_WIDTH = 1100
WINDOW_HEIGHT = 800

RCHECKER_WIDTH = 1000
RCHECKER_HEIGHT = 700

PROCEXPLORER_WIDTH = 640
PROCEXPLORER_HEIGHT = 580

def actionNameFromSystemName(name):
    return "SelectSystem-%s" % name

AUTHOR = "Toby Murray <toby.murray@comlab.ox.ac.uk>"
APPNAME = "Authodox"
VERSION = "0.2.0"
AUTHORS = [AUTHOR]

LIBEXEC_DIR = "/libexec/"
LOGO_FILENAME = "authodox-logo.tiff"
PREDICATES_FILENAME = "predicates.xml"
UI_FILENAME = "menu.xml"

DOC_DIR = "/doc/"
HELP_FILENAME = "index.xml"

def makeFileChooser(title, action, button, filterList):
    
    fileChooser = gtk.FileChooserDialog(title,
                                        None,
                                        action,
                                        (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                         button, gtk.RESPONSE_OK))
    fileChooser.set_default_response(gtk.RESPONSE_OK)
    for fil in filterList:
        filter = gtk.FileFilter()
        filter.set_name(fil[0])
        filter.add_pattern(fil[1])
        fileChooser.add_filter(filter)
        
    filter = gtk.FileFilter()
    filter.set_name("All files")
    filter.add_pattern("*")
    fileChooser.add_filter(filter)
    return fileChooser

def doSaveChangesBeforeClosingDialog(name):
    dialog = gtk.MessageDialog(None, gtk.DIALOG_MODAL, gtk.MESSAGE_WARNING,
                               gtk.BUTTONS_NONE,
                               "Save the changes to system \"%s\" before closing?" % name)
    dialog.format_secondary_text("If you don't save, changes will be permanently lost.")
    #dialog.set_markup("<span weight=\"bold\" size=\"large\">Save the changes to system \"%s\" before closing?</span>\n\n<span>If you don't save, changed will be permanently lost.</span>" % name)
    dialog.add_buttons("Close _without Saving", gtk.RESPONSE_CLOSE,
                       gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                       gtk.STOCK_SAVE, gtk.RESPONSE_OK)
    dialog.set_default_response(gtk.RESPONSE_OK)
    result = dialog.run()
    dialog.destroy()
    return result

def doRevertConfirmationDialog(name):
    dialog = gtk.MessageDialog(None, gtk.DIALOG_MODAL, gtk.MESSAGE_QUESTION,
                               gtk.BUTTONS_NONE,
                               "Revert unsaved changes to system \"%s\"?" % name)
    dialog.format_secondary_text("If you revert, changes will be permanently lost.")

    dialog.add_buttons(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                       gtk.STOCK_REVERT_TO_SAVED, gtk.RESPONSE_OK)
    dialog.set_default_response(gtk.RESPONSE_CANCEL)
    result = dialog.run()
    dialog.destroy()
    return result

def showErrorDialog(msg1, msg2 = None):
    dialog = gtk.MessageDialog(None, gtk.DIALOG_MODAL, gtk.MESSAGE_ERROR,
                               gtk.BUTTONS_OK,
                               msg1)
    if msg2 != None:
        dialog.format_secondary_text(msg2)
    dialog.run()
    dialog.destroy()

    
class OpenedSystem(object):

    def __init__(self, name, editor, hasBackingStore, systemsMenuItem, logWindow, dox):
        self.name = name
        self.editor = editor
        self.hasBackingStore = hasBackingStore
        self.systemsMenuItem = systemsMenuItem
        self.dox = dox
        editor.system.addModifiedCallback(self.systemIsModified)
        editor.system.addSyncedCallback(self.systemIsSynced)
        editor.addModifiedCallback(self.editorIsModified)
        editor.addSyncedCallback(self.editorIsSynced)   
        
        self.logWindow = logWindow
        vpane = gtk.VPaned()
        vpane.pack1(self.editor.getWidget(), True)
        vpane.pack2(self.logWindow.getWidget(), False)
        self.widget = vpane

    def systemIsModified(self, system):
        assert self.dox.currentSystem == self.name
        assert system == self.editor.system
        if self.hasBackingStore:
            self.dox.saveAction.set_sensitive(True)
            self.dox.revertAction.set_sensitive(True)            
        

    def systemIsSynced(self, system):
        assert self.dox.currentSystem == self.name
        assert system == self.editor.system        
        self.dox.saveAction.set_sensitive(False)
        self.dox.revertAction.set_sensitive(False)        

    def editorIsModified(self, editor):
        assert self.dox.currentSystem == self.name
        assert editor == self.editor
        self.dox.commitEditsAction.set_sensitive(True)
        self.dox.revertEditsAction.set_sensitive(True)        
        self.dox.checkRefinementsAction.set_sensitive(False)
        self.dox.exploreSystemAction.set_sensitive(False)
        self.dox.exploreExpressionAction.set_sensitive(False)
        self.dox.detectExcessAuthorityAction.set_sensitive(False)        

    def editorIsSynced(self, editor):
        assert self.dox.currentSystem == self.name
        assert editor == self.editor
        self.dox.commitEditsAction.set_sensitive(False)
        self.dox.revertEditsAction.set_sensitive(False)        
        self.dox.checkRefinementsAction.set_sensitive(True)
        self.dox.exploreSystemAction.set_sensitive(True)
        self.dox.exploreExpressionAction.set_sensitive(True)
        self.dox.detectExcessAuthorityAction.set_sensitive(True)

        
# an object that implements the main authodox window
class Dox(object):

    def deleteEvent(self, widget, event, data=None):
        return not self.quit()

    def destroy(self, widget, data=None):
        # quit on destroy
        gtk.main_quit()
        
    def quitCallback(self, action):
        if self.quit():
            # destroy the main window
            self.window.destroy()

    def quit(self):
        # close all of the files that are open, then quit
        for name in self.systems.keys():
            if not self.closeFile(name):
                return False
            
        return True

    def __init__(self):
        if "AUTHODOXHOME" not in os.environ:
            raise EnvironmentError("AUTHODOXHOME environment variable not set.")

        self.homeDir = os.path.abspath(os.environ["AUTHODOXHOME"])
        self.resourceDir = self.homeDir + LIBEXEC_DIR
        self.docDir = self.homeDir + DOC_DIR
        self.logoFilename = self.resourceDir + LOGO_FILENAME
        self.predicatesFilename = self.resourceDir + PREDICATES_FILENAME
        self.uiFilename = self.resourceDir + UI_FILENAME

        self.predicates = predicates.Predicates(self.predicatesFilename)
        
        self.logoPixbuf = gtk.gdk.pixbuf_new_from_file(self.logoFilename)
        
        gtk.window_set_default_icon(self.logoPixbuf)
        
        self.fdr = fdr.FDR()
        self.currentSystem = None
        self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
        self.window.set_title(APPNAME)
        self.window.connect("destroy", self.destroy)
        self.window.connect("delete_event", self.deleteEvent)
        self.window.set_default_size(WINDOW_WIDTH,WINDOW_HEIGHT)
        self.vbox = gtk.VBox()
        self.vbox.set_spacing(2)
        self.uiManager = gtk.UIManager()
        
        self.actionGroup = gtk.ActionGroup("DoxActions")
             # name           stock-id       label   accel tooltip callback
        self.actions = [
          ("FileAction",     None,          "_File"),
          ("ThisSystemAction",  None,       "_System"),          
          ("SystemsAction",  None,          "_Windows"),
          ("HelpAction",     None,          "_Help"),
          ("ContentsHelpAction",gtk.STOCK_HELP,"_Contents", "F1", None, self.contentsHelpCallback),
          ("AboutHelpAction",gtk.STOCK_ABOUT,"_About", None, None, self.aboutHelpCallback),
          ("NewFileAction",  gtk.STOCK_NEW, "_New", None, None, self.newFileCallback),
          ("OpenFileAction", gtk.STOCK_OPEN,"_Open",None, None, self.openFileCallback),
          ("SaveFileAction", gtk.STOCK_SAVE,"_Save",None, None, self.saveFileCallback),
          ("SaveAsFileAction", gtk.STOCK_SAVE_AS,"Save _As",None, None, self.saveAsFileCallback),
          ("ExportToCSPFileAction", None,"_Export to CSP",None, None, self.exportToCSPCallback),
          ("RevertFileAction", gtk.STOCK_REVERT_TO_SAVED,"_Revert",None, None, self.revertFileCallback),                    
          ("CloseFileAction",gtk.STOCK_CLOSE,"_Close",None, None,self.closeFileCallback),          
          ("QuitAction",     gtk.STOCK_QUIT,"_Quit",None, None, self.quitCallback),
          ("CheckRefinementsAction",     util.STOCK_CHECK,"Chec_k Refinements and Assertions","<Control>k", None, self.checkRefinementsCallback),
          ("DetectExcessAuthorityAction",     None ,"Detect Excess _Authority","<Control>d", None, self.detectExcessAuthorityCallback),          
          ("CommitChangesInEditorAction",     gtk.STOCK_APPLY,"_Commit Edits",None, None, self.commitEditsCallback),
          ("RevertChangesInEditorAction",     None,"_Revert Edits",None, None, self.revertEditsCallback),          
          ("ExploreSystemAction", util.STOCK_EXPLORE,"_Explore System","<Control>e", None, self.exploreSystemCallback),
          ("ExploreExpressionAction", util.STOCK_EXPLORE,"Explore","", None, self.exploreExpressionCallback)                                        
           ]
        
        self.actionGroup.add_actions(self.actions)
        self.closeAction = self.actionGroup.get_action("CloseFileAction")
        self.saveAction = self.actionGroup.get_action("SaveFileAction")
        self.saveAsAction = self.actionGroup.get_action("SaveAsFileAction")
        self.commitEditsAction = self.actionGroup.get_action("CommitChangesInEditorAction")
        self.revertEditsAction = self.actionGroup.get_action("RevertChangesInEditorAction")        
        self.checkRefinementsAction = self.actionGroup.get_action("CheckRefinementsAction")
        self.exportToCSPAction = self.actionGroup.get_action("ExportToCSPFileAction")
        self.exploreSystemAction = self.actionGroup.get_action("ExploreSystemAction")
        self.exploreExpressionAction = self.actionGroup.get_action("ExploreExpressionAction")        
        self.revertAction = self.actionGroup.get_action("RevertFileAction")

        self.detectExcessAuthorityAction = self.actionGroup.get_action("DetectExcessAuthorityAction")                

        
        for action in (self.closeAction, self.saveAction, self.saveAsAction,
                       self.commitEditsAction, self.revertEditsAction,
                       self.checkRefinementsAction,
                       self.exportToCSPAction, self.exploreSystemAction,
                       self.revertAction, self.exploreExpressionAction,
                       self.detectExcessAuthorityAction):
            action.set_sensitive(False)
            
        self.uiManager.insert_action_group(self.actionGroup, 0)
        self.uiManager.add_ui_from_file(self.uiFilename)
        self.systemsMenu = self.uiManager.get_widget("/MenuBar/SystemsMenu")
        self.systemsMenu.show()
        self.systemsMenu = self.systemsMenu.get_submenu()
        children = self.systemsMenu.get_children()
        for child in children:
            self.systemsMenu.remove(child)

        # set the icon for the detect excess authority toolbutton
        detectExcessAuthorityToolitem = self.uiManager.get_widget("/ToolBar/DetectExcessAuthority")
        toolitemIconSize = gtk.icon_size_lookup(detectExcessAuthorityToolitem.get_icon_size())
        toolbarLogoPixbuf = gtk.gdk.pixbuf_new_from_file_at_size(self.logoFilename, toolitemIconSize[0], toolitemIconSize[1])
        self.toolbarLogoImage = gtk.Image()
        self.toolbarLogoImage.set_from_pixbuf(toolbarLogoPixbuf)
        self.toolbarLogoImage.show()
        detectExcessAuthorityToolitem.set_icon_widget(self.toolbarLogoImage)

        # set the image for the detect excess authority menuitem
        detectExcessAuthorityMenuitem = self.uiManager.get_widget("/MenuBar/ThisSystemMenu/DetectExcessAuthority")
        img = detectExcessAuthorityMenuitem.get_image()
        (stock, size) = img.get_stock()
        menuitemIconSize = gtk.icon_size_lookup(size)
        menuitemLogoPixbuf = gtk.gdk.pixbuf_new_from_file_at_size(self.logoFilename, menuitemIconSize[0], menuitemIconSize[1])
        img.set_from_pixbuf(menuitemLogoPixbuf)
        
        self.accelGroup = self.uiManager.get_accel_group()
        self.window.add_accel_group(self.accelGroup)
        self.window.add(self.vbox)
        self.menubar = self.uiManager.get_widget("/MenuBar")
        self.vbox.pack_start(self.menubar, False, False)
        self.toolbar = self.uiManager.get_widget("/ToolBar")
        self.vbox.pack_start(self.toolbar, False, False)
        
        self.expressionExplorerBox = gtk.HBox()
        self.expressionExplorerBox.set_spacing(util.HSPACING)
        self.expressionExplorerBox.pack_start(gtk.Label("Process Explorer: "), False, False, 0)
        self.expressionExplorerEntry = gtk.Entry()
        self.expressionExplorerButton = gtk.Button("_Explore")
        image = gtk.Image()
        image.set_from_stock(util.STOCK_EXPLORE, gtk.ICON_SIZE_MENU)
        self.expressionExplorerButton.set_property("image", image)
        self.exploreExpressionAction.connect_proxy(self.expressionExplorerButton)
        self.exploreExpressionAction.connect_proxy(self.expressionExplorerEntry)
        self.expressionExplorerEntry.connect("activate", self.expressionExplorerEntryActivated)
        
        self.expressionExplorerBox.pack_start(self.expressionExplorerEntry, True, True, 0)
        self.expressionExplorerBox.pack_start(self.expressionExplorerButton, False, False, 0)
        self.vbox.pack_start(self.expressionExplorerBox, False, False)
        self.vbox.pack_start(gtk.HSeparator(), False, False)
        
        self.notebook = gtk.Notebook()
        self.notebook.popup_enable()
        self.notebook.connect("switch-page", self.notebookPageSwitched)
        self.notebook.set_scrollable(True)
        self.vbox.pack_start(self.notebook, True, True)
        self.window.show_all()
        self.systems = {}
        self.newSystemCount = 1

    def notebookPageSwitched(self, notebook, page, pageNum):
        assert notebook == self.notebook
        child = notebook.get_nth_page(pageNum)
        name = notebook.get_tab_label(child).get_text()
        self.systemSelected(name)
        
    def selectSystemCallback(self, action, name):
        self.selectSystem(name)
        
    def selectSystem(self, name):
        assert name in self.systems.keys()

        pageNum = self.notebook.page_num(self.systems[name].widget)
        if  pageNum == -1:
            pageNum = self.notebook.append_page(self.systems[name].widget,
                                                gtk.Label(name))
        assert pageNum != -1

        self.notebook.set_current_page(pageNum)

    def systemSelected(self, name):
        openedSys = self.systems[name]
        editor = openedSys.editor

        self.currentSystem = name
        self.closeAction.set_sensitive(True)
        self.saveAsAction.set_sensitive(True)
        self.window.set_title("%s - %s" % (name, APPNAME))
        self.window.show_all()
        self.saveAction.set_sensitive(openedSys.hasBackingStore and openedSys.editor.system.getModified())
        self.revertAction.set_sensitive(openedSys.hasBackingStore and openedSys.editor.system.getModified())        
        modifiedEditor = openedSys.editor.getModified()
        self.commitEditsAction.set_sensitive(modifiedEditor)
        self.revertEditsAction.set_sensitive(modifiedEditor)        
        self.checkRefinementsAction.set_sensitive(not modifiedEditor)
        self.exportToCSPAction.set_sensitive(True)
        self.exploreSystemAction.set_sensitive(not modifiedEditor)
        self.exploreExpressionAction.set_sensitive(not modifiedEditor)
        self.detectExcessAuthorityAction.set_sensitive(not modifiedEditor)

    def editSystem(self, sys, name, hasBackingStore):
        editor = None
        if name not in self.systems.keys():
            logWindow = LogWindow("Log for %s" % name)
            editor = systemeditor.SystemEditor(sys, logWindow,
                                               lambda : self.window.show_all(),
                                               WINDOW_WIDTH, WINDOW_HEIGHT,
                                               lambda obj : self.exploreSystem(name,"behaviour(%s)" % obj),
                                               lambda assname : self.checkAssert(assname),
                                               self.predicates)
            actionName = actionNameFromSystemName(name)
            action = (actionName, None, name, None, None, self.selectSystemCallback)
            self.actionGroup.add_actions([action], name)
            action = self.actionGroup.get_action(actionName)
            action.set_accel_group(self.accelGroup)
            menuitem = action.create_menu_item()
            openedSystem = OpenedSystem(name, editor, hasBackingStore, menuitem, logWindow, self)
            
            self.systems[name] = openedSystem
            self.systemsMenu.append(menuitem)
            menuitem.show()

        self.selectSystem(name)
                
    def newFileCallback(self, action):
        self.newFile()

    def newFile(self):
        sys = system.System()
        name = "Unsaved System %s" % str(self.newSystemCount)
        self.newSystemCount += 1
        self.editSystem(sys, name, False)
        
    def openFileCallback(self, action):
        fileChooser = makeFileChooser("Open File",
                                      gtk.FILE_CHOOSER_ACTION_OPEN,
                                      gtk.STOCK_OPEN,
                                      (("XML Files", "*.xml"),))
        result = fileChooser.run()
        if result == gtk.RESPONSE_OK:
            filename = fileChooser.get_filename()
            fileChooser.destroy()
            self.openFile(filename)
        else:
            fileChooser.destroy()

    def openFile(self, filename):
        filename = os.path.abspath(filename)
        xmlFile = None
        try:
            # try opening the file
            try:
                xmlFile = open(filename)
            except IOError, e:
                xmlFile = None
                if e.errno != errno.ENOENT:
                    raise

            # xmlFile == None iff file doesn't exist. In this case,
            # create an empty system with the name of the file and mark
            # the system as modified straightaway
            sys = system.System(xmlFile)
            if xmlFile != None:
                xmlFile.close()
                xmlFile = None
            else:
                sys.modified()
                
            self.editSystem(sys, filename, True)
                

        except Exception, e:
            showErrorDialog("Error opening file %s." % filename, str(e))
            if xmlFile != None:
                xmlFile.close()
        
    def saveFileCallback(self, action):
        assert self.currentSystem != None
        assert action == self.saveAction
        
        self.saveFile(self.currentSystem)

    # forces the user to sync the editor before continuing
    # returns True if the user chose to sync their edits and the sync
    # occurred successfully. Returns False otherwise
    def doSyncEditsDialog(self, name):
        dialog = gtk.MessageDialog(None, gtk.DIALOG_MODAL, gtk.MESSAGE_WARNING,
                                   gtk.BUTTONS_NONE,
                                   "System \"%s\" has uncommitted modifications." % name)
        dialog.format_secondary_text("All edits should be committed or reverted before proceeding.")
        dialog.add_buttons(gtk.STOCK_CANCEL, 0,
                           "_Revert All Edits and Continue", 1,
                           "_Commit All Edits and Continue", 2)
        dialog.set_default_response(2)
        result = dialog.run()
        dialog.destroy()
        if result == 2:
            return self.commitEdits(name)            
        elif result == 1:
            return self.revertEdits(name)
        else:
            assert result == 0
        return False
        
    def saveFile(self, name):
        def do():
            xmlFile = None
            try:
                openedSys = self.systems[name]
                assert openedSys.hasBackingStore
                filename = name
                xmlFile = open(filename, "w")
                openedSys.editor.system.toXML(xmlFile)
                xmlFile.close()
                xmlFile = None
            finally:
                if xmlFile != None:
                    xmlFile.close()

        if self.systems[name].editor.getModified():
            if not self.doSyncEditsDialog(name):
                return False
        assert not self.systems[name].editor.getModified()
        
        return self.systems[name].logWindow.performAndLog(do,(),"Saving file %s" % name)

    def saveAsFileCallback(self, action):
        assert self.currentSystem != None
        assert action == self.saveAsAction

        try:
            self.saveAsFile(self.currentSystem)
        except Exception, e:
            showErrorDialog("Error saving %s." % self.currentSystem, str(e))
        
    # returns True if successful, False otherwise
    # if saveOnly is True, then we don't close the old file and open the new
    # after the save has happened which is useful when we're shutting down
    def saveAsFile(self, name, saveOnly = False):
        if self.systems[name].editor.getModified():
            if not self.doSyncEditsDialog(name):
                return False
        assert not self.systems[name].editor.getModified()
        
        ret = False
        fileChooser = makeFileChooser("Save As",
                                      gtk.FILE_CHOOSER_ACTION_SAVE,
                                      gtk.STOCK_SAVE_AS,
                                      (("XML Files", "*.xml"),))
        result = fileChooser.run()
        if result == gtk.RESPONSE_OK:
            filename = fileChooser.get_filename()
            fileChooser.destroy()            
            xmlFile = None
            try:
                xmlFile = open(filename,"w")
                self.systems[name].editor.system.toXML(xmlFile)
                xmlFile.close()
                xmlFile = None
            finally:
                if xmlFile != None:
                    xmlFile.close()
                    xmlFile = None
                    
            if not saveOnly:
                xmlFile = None
                try:
                    result = self.closeFile(name)
                    assert result == True
                    self.currentSystem = None

                    xmlFile = open(filename)
                    sys = system.System(filename)
                    xmlFile.close()
                    xmlFile = None
                    self.editSystem(sys, filename, True)
                finally:
                    if xmlFile != None:
                        xmlFile.close()
                        xmlFile = None
                
            ret = True
        else:
            fileChooser.destroy()            

        return ret
        

    def closeFileCallback(self, closeAction):
        assert self.currentSystem != None
        assert closeAction == self.closeAction

        self.closeFile(self.currentSystem)

    # returns True if closed successfully, False otherwise (e.g. user
    # might have cancelled the close)
    def closeFile(self, name, force = False):
        result = False
        openedSys = self.systems[name]

        if not force:
            if openedSys.editor.getModified():
                if not self.doSyncEditsDialog(name):
                    return False
            assert not openedSys.editor.getModified()
        
            if openedSys.editor.system.getModified():
                response = doSaveChangesBeforeClosingDialog(name)
                # check the response, either bail (gtk.RESPONSE_CANCEL),
                # save or save-as (gtk.RESPONSE_OK) or 
                # or continue without saving (gtk.RESPONSE_CLOSE)
                if response == gtk.RESPONSE_CANCEL or response == gtk.RESPONSE_DELETE_EVENT:
                    return result
                elif response == gtk.RESPONSE_OK:
                    # save or save-as the bugger, if that works then continue
                    if openedSys.hasBackingStore:
                        if not self.saveFile(name):
                            return result
                    else:
                        if not self.saveAsFile(name, True):
                            return result
                
                        
                else:
                    assert response == gtk.RESPONSE_CLOSE
                    # continue
        
        # remove the action associated with this system
        action = self.actionGroup.get_action(actionNameFromSystemName(name))
        self.actionGroup.remove_action(action)

        # remove the menu item
        self.systemsMenu.remove(openedSys.systemsMenuItem)

        widget = openedSys.widget
        pageNum = self.notebook.page_num(widget)
        assert pageNum != -1
        self.notebook.remove_page(pageNum)
        
        del self.systems[name]

        if len(self.systems.keys()) == 0:
            for action in (self.closeAction, self.saveAsAction,
                           self.saveAction, self.commitEditsAction,
                           self.revertEditsAction,
                           self.checkRefinementsAction,
                           self.exportToCSPAction,
                           self.exploreSystemAction,
                           self.exploreExpressionAction,
                           self.revertAction,
                           self.detectExcessAuthorityAction):
                action.set_sensitive(False)
            self.window.set_title("%s" % APPNAME)
            self.currentSystem = None

        result = True
        return result

    def aboutHelpCallback(self, action):
        dialog = gtk.AboutDialog()
        dialog.set_name(APPNAME)
        dialog.set_version(VERSION)
        dialog.set_authors(AUTHORS)
        dialog.set_logo(self.logoPixbuf)
        dialog.run()
        dialog.destroy()

    def contentsHelpCallback(self, action):
        gnome.help_display_uri("ghelp:" + self.docDir + HELP_FILENAME)

    def commitEditsCallback(self, action):
        assert self.currentSystem != None
        assert action == self.commitEditsAction
        
        self.commitEdits(self.currentSystem)

    def commitEdits(self, name):
        editor = self.systems[name].editor
        return self.systems[name].logWindow.performAndLog(editor.commit,(),"Committing edits")

    def revertEditsCallback(self, action):
        assert self.currentSystem != None
        assert action == self.revertEditsAction
        
        self.revertEdits(self.currentSystem)
        
    def revertEdits(self, name):
        editor = self.systems[name].editor
        return self.systems[name].logWindow.performAndLog(editor.revert,(),"Reverting edits")
                
    def checkRefinementsCallback(self, action):
        assert self.currentSystem != None
        assert action == self.checkRefinementsAction
        
        self.checkRefinements(self.currentSystem)

    def checkAssert(self, assname):
        assert self.currentSystem != None
        name = self.currentSystem
        openedSys = self.systems[name]
        def do():
            refinements = {}
            preconds = ()
            code = ""
        
            assertion = openedSys.editor.system.assertions[assname]
            (assertionsRefinements, code) = self.predicates.buildConcreteRefinementAsserts((assertion,))
            (precond, refinement, cexViewer) = assertionsRefinements[assertion]
            if precond != None:
                preconds += (precond,)

            refinements[refinement] = (assname, cexViewer)
        
            # export to CSP, include the code from the refinements
            # this also generates concrete behaviour refinement asserts
            tmpFileName = "/tmp/authodox-%d.csp" % os.getpid()
            res = self.exportToCSP(name, tmpFileName, code)
        
            if res == False:
                raise Exception("Error exporting system to CSP")

            (val, linesToSections) = res
            assert val == True
        
            # check the preconditions
            result = openedSys.logWindow.performAndLog(
                self.fdr.checkAsserts,
                (tmpFileName, preconds, linesToSections),
                "Checking preconditions of assertions",
                True)
            
            if result == False:
                raise Exception("Error checking preconditions of assertions.")
            else:
                (value, res) = result
                assert value == True
                assert len(res) == len(preconds)
                if res != (True,)*len(res):
                    errorString = "Precondition failure.\n"
                    errorString += "%s is false when it should be true.\n" % preconds[0]
                    raise Exception(errorString)
            

                desc = "Checking Assertion %s in %s" % (assname,name)
                rChecker = refinementchecker.RefinementChecker(
                    "Checking Assertion %s" % assname,
                    desc,
                    self.fdr,
                    refinements,
                    tmpFileName,
                    linesToSections)
                window = gtk.Window(gtk.WINDOW_TOPLEVEL)
            
                def deleteEvent(widget, event, data=None):
                    assert widget == window
                    if not rChecker.isAlive():
                        rChecker.join()
                        return False
                    else:
                        return True
                
                window.connect("delete_event",deleteEvent)
                window.set_title(desc)
                window.add(rChecker.getWidget())
                window.set_default_size(RCHECKER_WIDTH,RCHECKER_HEIGHT)
                window.show_all()
                rChecker.start()
            
        return openedSys.logWindow.performAndLog(do,(),
                                                 "Checking assertion %s" % assname)        
        
    def checkRefinements(self, name):
        openedSys = self.systems[name]
        def do():

            refinements = {}
            preconds = ()
            code = ""
            
            assertions = ()
            for assname in openedSys.editor.system.assertions.keys():
                assertion = openedSys.editor.system.assertions[assname]
                assertions += (assertion,)

            (assertionsRefinements, code) = self.predicates.buildConcreteRefinementAsserts(assertions)
            
            for assname in openedSys.editor.system.assertions.keys():
                assertion = openedSys.editor.system.assertions[assname]
                (precond, refinement, cexViewer) = assertionsRefinements[assertion]
                if precond != None:
                    preconds += (precond,)
                refinements[refinement] = (assname, cexViewer)


            
            # export to CSP, include the code from the refinements
            # this also generates concrete behaviour refinement asserts
            tmpFileName = "/tmp/authodox-%d.csp" % os.getpid()
            res = self.exportToCSP(name, tmpFileName, code)
            
            if res == False:
                raise Exception("Error exporting system to CSP")

            (val, linesToSections) = res
            assert val == True

            assertions = openedSys.editor.system.concreteBehaviourAsserts
            for ass in assertions:
                assString = "%s refines %s" % (ass[1], ass[0])
                refinements[ass] = (assString, None)
                
            # check the preconditions
            result = openedSys.logWindow.performAndLog(
                self.fdr.checkAsserts,
                (tmpFileName, preconds, linesToSections),
                "Checking preconditions of assertions",
                True)
            
            if result == False:
                raise Exception("Error checking preconditions of assertions.")
            else:
                (value, res) = result
                assert value == True
                assert len(res) == len(preconds)
                if res != (True,) * len(preconds):
                    errorString = "Precondition failure.\n"
                    for index in range(len(res)):
                        if not res[index]:
                            errorString += "%s is false when it should be true.\n" % preconds[index]
                    raise Exception(errorString)
            

            desc = "Checking Refinements and Assertions of System in %s" % name
            rChecker = refinementchecker.RefinementChecker(
                "Checking Refinements and Assertions",
                desc,
                self.fdr,
                refinements,
                tmpFileName,
                linesToSections)
            window = gtk.Window(gtk.WINDOW_TOPLEVEL)
            
            def deleteEvent(widget, event, data=None):
                assert widget == window
                if not rChecker.isAlive():
                    rChecker.join()
                    return False
                else:
                    return True
                
            window.connect("delete_event",deleteEvent)
            window.set_title(desc)
            window.add(rChecker.getWidget())
            window.set_default_size(RCHECKER_WIDTH,RCHECKER_HEIGHT)
            window.show_all()
            rChecker.start()
            
        return openedSys.logWindow.performAndLog(do,(),
                                                 "Checking refinements and assertions")        

    def exportToCSPCallback(self, action):
        assert self.currentSystem != None
        assert action == self.exportToCSPAction
        
        name = self.currentSystem
        if self.systems[name].editor.getModified():
            if not self.doSyncEditsDialog(name):
                return

        fileChooser = makeFileChooser("Export to CSP",
                                      gtk.FILE_CHOOSER_ACTION_SAVE,
                                      "_Export",
                                      (("CSP Files", "*.csp"),))
            
        result = fileChooser.run()
        if result == gtk.RESPONSE_OK:
            filename = fileChooser.get_filename()
            fileChooser.destroy()
            self.exportToCSP(self.currentSystem, filename)
        else:
            fileChooser.destroy()
        

    # returns False on error
    # (True, linesToSections) on success
    def exportToCSP(self, name, filename, extra = None):
        openedSys = self.systems[name]
        assert not openedSys.editor.getModified()
        
        def do():
            if not openedSys.logWindow.performAndLog(openedSys.editor.system.buildConcreteBehaviourAsserts,(),"Building behaviour refinement assertions"):
                raise Exception("Error building behaviour refinement assertions.")
                
            def do():
                cspFile = None
                linesToSections = None
                
                try:
                    cspFile = open(filename,"w")
                    linesToSections = openedSys.editor.system.exportToCSP(cspFile)
                    if extra != None:
                        cspFile.write(extra)
                    cspFile.close()
                    cspFile = None
                finally:
                    if cspFile != None:
                        cspFile.close()

                return linesToSections
            
            result = openedSys.logWindow.performAndLog(do,(),
                                                     "Exporting system to CSP file %s" % filename, True)
            if result == False:
                raise Exception("Error exporting to CSP file.")
            (value, res) = result
            assert value == True
            return res
        
        return openedSys.logWindow.performAndLog(do,(),
                                                 "Exporting system %s to CSP" % name, True)

    def exploreSystemCallback(self, action):
        assert self.currentSystem != None
        assert action == self.exploreSystemAction

        self.exploreSystem(self.currentSystem, "System")

    def exploreExpressionCallback(self, action):
        assert self.currentSystem != None
        assert action == self.exploreExpressionAction

        self.exploreSystem(self.currentSystem, self.expressionExplorerEntry.get_text())

    def expressionExplorerEntryActivated(self, entry):
        assert self.currentSystem != None
        assert entry == self.expressionExplorerEntry
        assert self.expressionExplorerButton.get_property("sensitive")

        self.expressionExplorerButton.clicked()

    def exploreSystem(self, name, process):
        openedSys = self.systems[name]
        def do():
            tmpFileName = "/tmp/authodox-%d.csp" % os.getpid()
            result = self.exportToCSP(name, tmpFileName)
            if result == False:
                raise Exception("Error exporting system to CSP")

            (value, linesToSections) = result
            assert value == True


            window = gtk.Window(gtk.WINDOW_TOPLEVEL)
            
            procExplorer = processexplorer.ProcessExplorer(
                self.fdr,
                tmpFileName,
                process,
                name,
                linesToSections,
                lambda : window.show_all())

            window.set_title(procExplorer.getDesc())
            window.add(procExplorer.getWidget())
            window.set_default_size(PROCEXPLORER_WIDTH,PROCEXPLORER_HEIGHT)
            
            def deleteEvent(widget, event, data=None):
                assert widget == window
                if not procExplorer.isAlive():
                    procExplorer.join()
                    return False
                else:
                    return True
            
            window.connect("delete_event",deleteEvent)
            window.show_all()
            
            procExplorer.start()
            
        return openedSys.logWindow.performAndLog(do,(),
                                                 "Exploring %s" % process)
    def revertFileCallback(self, action):
        assert self.currentSystem != None
        assert action == self.revertAction

        if doRevertConfirmationDialog(self.currentSystem) == gtk.RESPONSE_OK:
            self.revertFile(self.currentSystem)

    def revertFile(self, name):
        # force the close to occur
        assert self.systems[name].hasBackingStore
        
        result = self.closeFile(name, True)
        assert result

        try:
            filename = name
            xmlFile = open(filename)
            sys = system.System(filename)
            xmlFile.close()
            xmlFile = None
            self.editSystem(sys, filename, True)
            return True
        except Exception, e:
            showErrorDialog("Error opening file %s." % filename, str(e))
        
        if xmlFile != None:
            xmlFile.close()
        return False
        
    def detectExcessAuthorityCallback(self, action):
        assert action == self.detectExcessAuthorityAction
        assert self.currentSystem != None

        # detect excess authority for all of the objects that have
        # some allowed authority specified for them in the current system
        self.detectExcessAuthority(self.currentSystem, self.systems[self.currentSystem].editor.system.allowedAuthority.keys())

    # detect excess authority in each of the objects in 'objects' in the
    # the system 'name'
    def detectExcessAuthority(self, name, objects):
        openedSys = self.systems[name]

        preds = []
        predDescs = {}
        for obj in objects:
            authority = openedSys.editor.system.allowedAuthority[obj]
            if authority.perms != None:
                pred = ( "NP", ("diff({|ioe_.%s|},allowedPerms_(%s))" % (obj, obj),"System") )
                preds += [ pred ]
                predDescs[pred] = "%s can't acquire permissions in excess of %s" % (obj, authority.perms)

            if authority.wpc != None:
                pred = ( "RCP_NTCp", ("alphabet(%s)" % obj, "diff(theAlphabet,union(allowedWPC_(%s),alphabet(%s)))" % (obj, obj),"System") )
                preds += [ pred ]
                predDescs[pred] = "%s can't weakly cause any event to occur outside of %s" % (obj, authority.wpc)

            if authority.spc != None:
                pred = ( "RCNTCp", ("alphabet(%s)" % obj, "diff(theAlphabet,union(allowedSPC_(%s),alphabet(%s)))" % (obj, obj),"System") )
                preds += [ pred ]
                predDescs[pred] = "%s can't strongly cause any event to occur outside of %s" % (obj, authority.spc)

            if authority.wnc != None:
                pred = ( "RCP_NTCn", ("alphabet(%s)" % obj, "diff(theAlphabet,union(allowedWNC_(%s),alphabet(%s)))" % (obj, obj),"System") )
                preds += [ pred ]
                predDescs[pred] = "%s can't weakly cause any event to not occur outside of %s" % (obj, authority.wnc)

            if authority.snc != None:
                pred = ( "RCNTCn", ("alphabet(%s)" % obj, "diff(theAlphabet,union(allowedSNC_(%s),alphabet(%s)))" % (obj, obj),"System") )
                preds += [ pred ]
                predDescs[pred] = "%s can't strongly cause any event to not occur outside of %s" % (obj, authority.snc)
                
                
            #if authority.dpc != None:
            #    pred = ( "RCDNTCp", ("alphabet(%s)" % obj, "diff(theAlphabet,union(allowedDPC_(%s),alphabet(%s)))" % (obj, obj),"System") )
            #    preds += [ pred ]
            #    predDescs[pred] = "%s can't directly cause any event to occur outside of %s" % (obj, authority.dpc)
                
        # turn the predicates into a concrete set of tests
        (dict, code) = self.predicates.buildConcreteRefinementAsserts(preds)

        asserts = []
        for v in dict.values():
            precond = v[0]
            if precond != None:
                asserts.append(precond)
        
        def do():
            tmpFileName = "/tmp/authodox-%d.csp" % os.getpid()
            result = self.exportToCSP(name, tmpFileName, code)
            if result == False:
                raise Exception("Error exporting system to CSP")
            (value, linesToSection) = result
            assert value == True
            
            result = openedSys.logWindow.performAndLog(
                self.fdr.checkAsserts,
                (tmpFileName, asserts, linesToSection),
                "Checking preconditions of assertions",
                True)
            
            if result == False:
                raise Exception("Error checking preconditions of assertions.")
            else:
                (value, res) = result
                assert value == True
                assert len(res) == len(asserts)
                if res != (True,) * len(asserts):
                    errorString = "Precondition failure in excess authority assertions.\n"
                    for index in range(len(res)):
                        if not res[index]:
                            errorString += "%s is false when it should be true.\n" % asserts[index]
                    raise Exception(errorString)

            refinements = {}
            for pred in preds:
                ref = dict[pred][1]
                desc = predDescs[pred]
                refinements[ref] = (desc, dict[pred][2])
            
            desc = "Detecting Excess Authority in %s" % name
            rChecker = refinementchecker.RefinementChecker(
                "Detecting Excess Authority",
                desc,
                self.fdr,
                refinements,
                tmpFileName,
                linesToSection)
            window = gtk.Window(gtk.WINDOW_TOPLEVEL)

            def deleteEvent(widget, event, data=None):
                assert widget == window
                if not rChecker.isAlive():
                    rChecker.join()
                    return False
                else:
                    return True
                
            window.connect("delete_event",deleteEvent)
            
            window.set_title(desc)
            window.add(rChecker.getWidget())
            window.set_default_size(RCHECKER_WIDTH,RCHECKER_HEIGHT)
            window.show_all()
            rChecker.start()
            
        return openedSys.logWindow.performAndLog(do,(),
                                                 "Detecting excess authority")
        
            
            
if __name__ == "__main__":
    gtk.threads_init()
    
    dox = Dox()
    args = sys.argv[1:]
    if len(args) == 0:
        dox.newFile()
    else:
        for arg in args:
            dox.openFile(arg)

    gtk.main()
