class EditorSyncError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return str(self.msg)

# a base class for UI components that edit things.
# we presume that the thing that an editor is editing, T, is separate from
# the editor. hence changing the value in the editor doesn't necessarily
# update the value in T. Hence, when the user edits the value, the editor
# might go into the "modified" state, indicating that its value is
# different from the last known value of T. Once the user decides to
# "commit" their edits, the editor becomes synchronised with T and so moves
# into the synchronised (i.e. unmodified) state.
#
# this base class proides facilities so that UI components can register
# callbacks to be notified of when an editor changes state, therby
# providing a common interface to editing UI sub-widgets
class Editor(object):
    # if modified is given, it indicates whether our initial status is modified
    def __init__(self, modified = False):
        self.modifiedCallbacks = []
        self.syncedCallbacks = []
        self.isModified = modified

    # this method is used to register a callback to be called
    # when the user modifies the view of the thing being edited
    # but the changes haven't been committed yet
    #
    # the given callback is passed the editor object when invoked
    def addModifiedCallback(self, callback):
        if callback not in self.modifiedCallbacks:
            self.modifiedCallbacks += [callback, ]

    # this method is used to register a callback to be called
    # when the user's view of the thing being edited is synchronised
    # with the thing itself (i.e. the user's changes are committed)
    #
    # the given callback is passed the editor object when invoked
    def addSyncedCallback(self, callback):
        assert callback not in self.syncedCallbacks
        self.syncedCallbacks += [callback, ]

    # this method used to unregister a modified callback
    def removeModifiedCallback(self, callback):
        assert callback in self.modifiedCallbacks
        self.modifiedCallbacks.remove(callback)

    # this method used to unregister a synced callback
    def removeSyncedCallback(self, callback):
        assert callback in self.syncedCallbacks
        self.syncedCallbacks.remove(callback)

    # calls all of the modified callbacks and sets the modified status True
    def modified(self):
        if not self.isModified:
            self.isModified = True
            for callback in self.modifiedCallbacks:
                callback(self)

    # calls all of the synced callbacks and sets the modified status False
    def synced(self):
        if self.isModified:
            self.isModified = False
            for callback in self.syncedCallbacks:
                callback(self)

    # returns the modified status
    def getModified(self):
        return self.isModified

    # forces the editor to revert its contents -- this synchronises
    # the editor with the underlying thing it is editing by pulling the
    # thing's contents to the editor.
    #
    # this method returns iff the synchronisation is successful
    # otherwise it raises a EditorSyncError
    # any implementation of this should always call self.synced()
    # before returning successfully
    def revert(self):
        raise NotImplementedError("Editor.revert() is abstract")
    
    # forces the editor to commit its contents -- this synchronises
    # the editor with the underlying thing it is editing by pushing its
    # modified contents to the underlying thing.
    #
    # this method returns iff the synchronisation is successful
    # otherwise it raises a EditorSyncError
    # any implementation of this should always call self.synced()
    # before returning successfully
    def commit(self):
        raise NotImplementedError("Editor.commit() is abstract")

    # get a gtk.Widget subclass that represents the UI of this editor
    def getWidget(self):
        raise NotImplementedError("Editor.getWidget() is abstract")

    

