# functions for dealing with simple expressions of the form
# "X(y1,y2,...,.yn)", where each X and yi is an identifier
# these are used in a couple of different places, such as for
# the names of abstract behaviours and the names of predicates
import re
import util

# regular expression to match a string of the form X(y1,...,yn) or just X
# these are things like CSP process names and predicates
regexp = r"""^                 # start of string
             (\w+'*)           # match the name
             (
              \s*              # whitespace
              \(               # opening bracket
              \s*              # whitespace
              (\s*(\w+'*)\s*,)*   # zero or more variables
              \s*              # whitespace
              (\w+'*)          # a final variable
              \s*              # whitespace
              \)               # closing bracket
             ){0,1}            # want either no or 1 paramlist
             $"""              # end of string

pattern = re.compile(regexp,re.VERBOSE)

class MatchError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return str(self.msg)

# check whether the given string is a valid expr
# return True if so, False otherwise
def isValid(str):
    m = pattern.match(str)
    return m != None

    
# parse an expression of the form "X(x1,x2,...,xn)" and return a tuple
# ("X",("x1","x2",...,"xn"))
# raises a MatchError if the string is not of correct form
def parse(str):
    m = pattern.match(str)
    if m == None:
        raise MatchError("String not of the correct form")

    g = m.groups()
    
    name = g[0]

    params = g[1]
    if params != None:
        # grab each param name
        rexp = r"\w+'*"
        pat = re.compile(rexp)
        params = tuple(pat.findall(params))

    return (name, params)


# given an expression of the form "X(x1,x2,...,xn)" and
# a tuple ("X", ("y1","y2",...,"yn")), return a dictionary of the form
# {"x1": "y1", "x2": "y2", ..., "xn": "yn"}
# if the expression is of the form "X" and the tuple is ("X",),
# an empty dictionary is returned
#
# raises a MatchError exception
def matchVars(stri, vals):

    (name1, params1) = parse(stri)
    (name2, params2) = vals

    dict = {}
    
    if name1 != name2:
        raise MatchError("Names do not match")

    if (params1 == None and params2 != None) or (params1 != None and params2 == None):
        raise MatchError("One expression has parameters while the other doesn't")
    if params1 != None:
        assert params2 != None
    
        if len(params1) != len(params2):
            raise MatchError("Different numbers of parameters")

        # add each to the dictionary
        for index in range(len(params1)):
            dict[params1[index]] = params2[index]

    return dict

# given a string of the form "X(y1,...,yn)" and a map containing
# {y1: x1, ..., yn: xn}, return the string
# "X(x1,...,xn)"
def replaceVars(str,map):
    match = pattern.match(str)
    assert match != None, str
    
    groups = match.groups()
    name = groups[0]
    
    result = name
    
    params = groups[1]

    if params != None:
        result += "("
        # grab each param name
        rexp = r"\w+'*"
        pat = re.compile(rexp)
    
        paramList = pat.findall(params)
        replacedList = []
        for param in paramList:
            if param in map:
                replacedList += [map[param]]
            else:
                replacedList += [param]
        result += util.seqString(replacedList, ", ")
        result += ")"
            
    return result



# look for identifiers in st. if they are mapped by map, replace them
# with the contents of map
def applyMapToStr(map, st):
    # we should get back a list of identifiers with the other
    # delimiting characters returned as elements of the list as well
    # -- this is why the regexp is surrounded by brackets, see the
    #    python library reference
    idents = re.split("(\w+'*)", st)
    result = []
    for ident in idents:
        if ident in map:
            result += map[ident]
        else:
            result += ident
    
    return util.seqString(result,"")
            
            
    
