import os
import signal
import popen2
import select
import re
import util
import time

def errorDuringNewSession(out, err):
    assert out == ""
    if err != "":
        return True
    return False

def errorDuringLoad(out, err):
    assert out == ""
    if err != " \x08Ready\n/\x08":
        return True
    return False

def errorDuringCompile(out, err):
    assert out == ""
    # a regular expression to match successful compilation output
    # to stderr by FDR
    regexp = r"""
    ^Starting\ timer$\n
    ^Starting\ compilation$\n^$\n
    ^Starting...$\n
    (^.*$\n)*    
    ^Compiling...$\n
    ^Reading...$\n
    (^.*$\n)*
    ^\ Loading...($\n^Ready$\n^){0,1}\ done$\n
    ^Took\ [0-9]+\([0-9]+\+[0-9]+\)\ seconds$"""
    
    if not re.search(regexp,err,re.VERBOSE | re.MULTILINE):
        return True
    return False

def errorDuringRefinedby(out, err):
    assert out == ""
    if err != "" and err != "\nReady\n":
        return True
    return False

def errorDuringAssrt(out, err):
    if (out != "true\n" and out != "false\n" and
        out != "xtrue\n" and out != "xfalse\n"):
        return True
    return False

def errorDuringConfig(out, err):
    if err != "" or out != "":
        return True
    return False

class FDRCommandError(Exception):
    def __init__(self, msg, out, outErr, linesToSections):
        self.msg = msg
        self.out = out
        self.outErr = outErr
        self.linesToSections = linesToSections

    def __str__(self):
        output = ""
        if self.out == "":
            output += self.msg + "\n" + "Output from FDR:\n" + self.outErr
        elif self.outErr == "":
            output += self.msg + "\n" + "Output from FDR:\n" + self.out
        else:
            output += self.msg + "\n" + "Output from FDR:\n" + "(FDR output to stdout)\n" + self.out + "(FDR output to stderr)\n" + self.outErr

        output += "End of output from FDR.\n\n"
        if self.linesToSections != None:
            lines = self.getLineNumbers()
            for line in lines:
                if line in self.linesToSections.keys():
                    output += "This error pertains to %s\n" % self.linesToSections[line]
        return output
        
    # return a tuple of line numbers mentioned in the error
    def getLineNumbers(self):
        msg = self.out + " " +  self.outErr
        regexp = "line\ [0-9][0-9]*\ of\ (?!standard\ input)"
        matches = re.findall(regexp, msg, re.IGNORECASE)
        lines = ()
        for match in matches:
            line = int(match.split(' ')[1])
            lines += (line, )

        return lines
        
# a single behaviour in a counterexample
class CounterExBehaviour(object):
    def __init__(self, example, actor, str):
        def parseSequence(seq, begin, end):
            assert seq.startswith(begin)
            seq = seq[len(begin):]
            assert seq.endswith(end)
            seq = seq[:len(seq)-len(end)]
            return seq.split(",")
        
        self.example = example
        self.actor = actor
        self.trace = None
        self.acceptances = None
        self.refusals = None
        str = str.split("\n")
        for line in str:
            if line.startswith("Performs "):
                line =  line.replace("Performs ","",1)
                self.trace = parseSequence(line, "<", ">")
            if line.startswith("Accepts "):
                line =  line.replace("Accepts ","",1)                
                self.acceptances = parseSequence(line, "{", "}")
            if line.startswith("Refuses "):
                line =  line.replace("Refuses ","",1)                
                self.refusals = parseSequence(line, "{", "}")
            

    def __str__(self):
        res = ("BEGIN BEHAVIOUR example=%s process=%s\n" %
               (self.example, self.actor))
        
        res += "BEGIN TRACE\n"
        res += str(self.trace) + "\n"
        res += "END TRACE\n"

        if (self.acceptances != None):
            res += "BEGIN ACCEPTANCES\n"
            res += str(self.acceptances) + "\n"
            res += "END ACCEPTANCES\n"

        if (self.refusals != None):
            res += "BEGIN REFUSALS\n"
            res += str(self.refusals) + "\n"
            res += "END REFUSALS\n"

        res += "END BEHAVIOUR\n"
        return res
    
# counterexamples to a failed assertion checked by FDR
class CounterExamples(object):
    
    # given an FDRInstance and the name of a failed assertion,
    # make a new CounterExamples object that represents the counterexamples
    # to that failed assertion
    def __init__(self, instance, assName):
        self.examples = []

        # create the debug object
        dbgName = "dbg" + assName
        (out, err) = instance.runCmd("%s debug %s\n" % (assName, dbgName))
        
        # get the number of counterexamples
        (out, err) = instance.runCmd("puts [%s witnesses]\n" % dbgName)
        witnesses = int(out)
        assert witnesses > 0
        
        # get the number of "actors" in these counterexamples
        (out, err) = instance.runCmd("puts [%s actors]\n" % dbgName)
        actors = int(out)

        # for each counterexample
        for witness in range(witnesses):
            thisExample = []

            # for each actor in each counterexample
            for actor in range(actors):

                # create a behaviour object for the actor
                behaviourName = dbgName + "behaviour" + str(witness) + "_" + str(actor)
                (out,err) = instance.runCmd("%s attribution %s %s %s\n" %
                                (dbgName, str(witness), str(actor),
                                 behaviourName))

                # get the description of this behaviour
                (out, err) = instance.runCmd("puts [%s describe]\n" %
                                             behaviourName)
                
                # cleanup the description somewhat
                out = out.replace("\x0c",",")
                out = out.replace("\x0b","\n")
                out = out.replace("\t"," ")

                thisExample += [CounterExBehaviour(witness,actor,out)]

            assert len(thisExample) == actors
            self.examples += [thisExample]

        assert len(self.examples) == witnesses


    # returns number of counterexamples
    def numExamples(self):
        return len(self.examples)

    # returns the number of actors in each counterexample
    def numActors(self):
        assert self.numExamples() > 0
        return len(self.examples[0])

    # convert to a string representation
    def __str__(self):
        res = ""
        for ex in range(self.numExamples()):
            res += "BEGIN COUNTEREXAMPLE %d\n\n" % ex
            for actor in range(self.numActors()):
                res += str(self.examples[ex][actor])
                res += "\n"
            res += "END COUNTEREXAMPLE\n\n\n"
        return res
                                      
        
# a running FDR instance
class FDRInstance(object):

    # create a new instance object given the file objects that
    # reprsent the FDR process's stdin, stdout and stderr respectively
    def __init__(self, toFDR, fromFDR, fromError, pid):
        self.toFDR = toFDR
        self.fromFDR = fromFDR
        self.fromError = fromError
        self.pid = pid

    # run the given command
    # returns a pair (out, err) where each is the output from the command
    # on stdout and stderr respectively
    #
    # if callbacks is not None, we expect it to be an object supporting two
    # methods: newFDROutput(str) and newFDRErrorOutput(str) which are called
    # when FDR produces new output on stdout stderr output respectively.
    def runCmd(self, cmd, callbacks = None):

            
        # ensure the command is EOL terminated
        cmd = util.eolTerminate(cmd)
        
        # ensure FDR's output includes messages at the beginning and
        # end of running the command -- this enables us to determine
        # when to stop reading output
        beginCmd = "puts \"beginCmd\"\n"
        endCmd = "puts \"endCmd\"\n"

        self.toFDR.write(beginCmd)
        self.toFDR.flush()
        
        line = self.fromFDR.readline()        
        if line != "beginCmd\n":
            raise IOError("Error reading from FDR")
        
        self.toFDR.write(cmd)
        self.toFDR.write(endCmd)
        self.toFDR.flush()

        #os.fsync(self.toFDR.fileno())

        fromfd = self.fromFDR.fileno()
        errfd = self.fromError.fileno()
        
        # read output until we see "endCmd"
        output = ""
        errOutput = ""
        callbackInterval = 0.3  # seconds

        errOutputSinceLastCall = ""
        outputSinceLastCall = ""
        timeOfLastCall = time.time()
        
        while not output.endswith("endCmd\n"):
            (rin, rout, rerr) = select.select([fromfd, errfd],
                                              [],
                                              [],
                                              callbackInterval)
            if errfd in rin:
                # read error output
                text = os.read(errfd, 4096)
                errOutput += text
                errOutputSinceLastCall += text
            if fromfd in rin:
                text = os.read(fromfd, 4096)
                output += text
                outputSinceLastCall += text                


            if callbacks != None:
                newTime = time.time()
                if newTime - timeOfLastCall >= callbackInterval:
                    timeOfLastCall = newTime
                    if outputSinceLastCall != "":
                        callbacks.newFDROutput(outputSinceLastCall)
                        outputSinceLastCall = ""
                    if errOutputSinceLastCall != "":
                        callbacks.newFDRErrOutput(errOutputSinceLastCall)
                        errOutputSinceLastCall = ""               

        if callbacks != None:
            if outputSinceLastCall != "":
                callbacks.newFDROutput(outputSinceLastCall)
            if errOutputSinceLastCall != "":
                callbacks.newFDRErrOutput(errOutputSinceLastCall)
                
        # string trailing "endCmd\n" from output
        output = output.replace("endCmd\n","")
        return (output, errOutput)

    # close the session -- exits the FDR instance and closes all descriptors
    def close(self):
        self.toFDR.write("exit\n")
        self.toFDR.flush()
        
        self.toFDR.close()
        self.fromFDR.close()
        self.fromError.close()
        
        # kill the child
        os.kill(self.pid, signal.SIGTERM)
        os.waitpid(self.pid, 0)
        
        self.toFDR = None
        self.fromFDR = None
        self.fromError = None
        


# a very basic class for using FDR
class FDR(object):

    VERBOSITY_AUTO = "auto"
    VERBOSITY_NONE = "none"
    VERBOSITY_MEDIUM = "medium"
    VERBOSITY_FULL = "full"
    
    # throws KeyError if FDRHOME environment variable is not defined
    def __init__(self):
        if "FDRHOME" not in os.environ:
            raise EnvironmentError("FDRHOME environment variable not set.")
        self.fdrhome = os.environ["FDRHOME"]
        self.bin = self.fdrhome + "/bin/fdr2tix"
        self.args = ["-insecure", "-nowindow"]
        self.cmdepilogue = "exit\n"
        self.cmdprologue = ("source " + self.fdrhome +
                            "/lib/fdr/fdrDirect.tcl\n")

    # start a new FDR instance -- returns an FDRInstance object if successful
    def newInstance(self):
        # launch fdr in command-line mode
        bin = [self.bin] + self.args
        child = popen2.Popen3(bin, True)
        


        instance = FDRInstance(child.tochild, child.fromchild, child.childerr, child.pid)
        instance.runCmd(self.cmdprologue)
        
        return instance
        
    # given a file name containing CSP code and a bunch of assertions
    # (boolean expressions),
    # load the code and check each of the assertions by having FDR
    # evaluate each one.
    #
    # returns a sequence of length equal to the original
    # sequence of asserts, the nth element of which indicates whether
    # the nth assertion was correct. False indicates a false assertion,
    # while True indicates the opposite.
    def checkAsserts(self, cspFileName, asserts, linesToSections = None):

        results = ()
        def runCmd(instance, command, errCheck):
            (out, err) = instance.runCmd(command)
            if errCheck(out,err):
                raise FDRCommandError("Error running FDR command: %s\n" % command, out, err, linesToSections)
            return (out, err)
        
        sessionName = "mysession"

        def newSession():
            return "session %s" % sessionName

        def load(fileName):
            (head, tail) = os.path.split(fileName)
            return "%s load %s %s" % (sessionName, head, tail)

        def errorDuringEvalAndPrint(out, err):
            if err != "" or (out != "true\n" and out != "false\n"):
                return True
            return False
        
        instance = self.newInstance()

        command = newSession()
        runCmd(instance, command, errorDuringNewSession)

        command = load(cspFileName)
        runCmd(instance, command, errorDuringLoad)

        for ass in asserts:
            command = "puts [mysession evaluate \"%s\"]" % ass
            (out, err) = runCmd(instance, command, errorDuringEvalAndPrint)

            if out == "true\n":
                results += (True,)
            else:
                assert out == "false\n"
                results += (False,)

        instance.close()
        return results
                
    # given a file name containing CSP code and a bunch of refinement asserts,
    # load the code and check each of the refinement asserts.
    # asserts is a sequence of pairs, (a, b), where each means
    # that we assert that a is refined by b
    #
    # returns a sequence of length equal to the original
    # sequence of asserts, the nth element of which indicates whether
    # the nth assertion was correct. None indicates a correct assertion.
    # If the nth assertion fails, a CounterExamples object is returned instead
    # of None
    #
    # if callbacks is not None, it is expected to be an object that has
    # the following methods: testingAssertion(assertionNum),
    #                        testedAssertion(assertionNum, result)
    #            as well as: newFDROutput(text),
    #                        newFDRErrOutput(text)
    #
    # the callbacks will be called to notify the caller of the progress
    def checkRefinements(self, cspFileName, asserts, callbacks = None,
                         maxCounterexamplesPerAssert = 1,
                         verbosity = VERBOSITY_AUTO,
                         linesToSections = None):
        result = ()
        
        # a mapping from real process names to simple autogenerated names
        fdrnames = {}
        
        instance = self.newInstance()
        output = ""
        errOutput = ""
        
        def runCmd(instance, command, errCheck):
            (out, err) = instance.runCmd(command, callbacks)
            if errCheck(out,err):
                raise FDRCommandError("Error running FDR command: %s\n" % command, out, err, linesToSections)
            return (out, err)
        
        def compileProc(self, proc):
            if proc not in fdrnames:
                fdrname = "proc" + str(len(fdrnames.keys()))
                fdrnames[proc] = fdrname
                command = compile(proc)
                runCmd(instance,command, errorDuringCompile)



        def config(param, value):
            return "config %s %s" % (param, value)

        sessionName = "mysession"

        def newSession():
            return "session %s" % sessionName

        def load(fileName):
            (head, tail) = os.path.split(fileName)
            return "%s load %s %s" % (sessionName, head, tail)

        def compile(proc):
            # there is a trick here that it took me a few hours to figure out
            # -- we have to get FDR to /evaluate/
            # the expression to be compiled before we compile the thing
            return ("%s compile [%s evaluate \"%s\"] %s" %
                    (sessionName, sessionName, proc, fdrnames[proc]))

        def refinedby(p1,p2,assName):
            return "%s refinedby %s %s" % (fdrnames[p1], fdrnames[p2],
                                           assName)

        def assrt(assName):
            return "puts [%s assert]" % assName
            
        command = config("max_errors", maxCounterexamplesPerAssert)
        runCmd(instance, command, errorDuringConfig)

        command = config("report", verbosity)
        runCmd(instance, command, errorDuringConfig)

        command = newSession()
        runCmd(instance, command, errorDuringNewSession)

        command = load(cspFileName)
        runCmd(instance, command, errorDuringLoad)

        assCount = 0
        for ass in asserts:

            if callbacks != None:
                callbacks.testingAssertion(assCount)
                
            assert(len(ass) == 2)
            p1 = ass[0]
            p2 = ass[1]            

            compileProc(self, p1)
            compileProc(self, p2)

            assName = "ass" + str(assCount)
            assCount += 1

            command = refinedby(p1, p2, assName)
            runCmd(instance, command, errorDuringRefinedby)

            command = assrt(assName)
            (out, err) = runCmd(instance,command, errorDuringAssrt)
            if out != "true\n" and out != "xtrue\n":
                res = CounterExamples(instance,assName)
            else:
                res = None

            result += (res,)
            if callbacks != None:
                callbacks.testedAssertion(assCount-1, res)
            
        instance.close()
        return result
    
    # Launch FDR and run the sequence of given commands before exiting.
    # We automatically 'source' fdrDirect.tcl so the builtins provided
    # by it are available.
    #
    # returns (output, outputError) -- the output from stdout and stderr resp.
    def runSession(self, commands):
        # launch fdr in command-line mode
        bin = [self.bin] + self.args
        (toChild, fromChild, fromChildError) = os.popen3(bin)

        toChild.writelines(self.cmdprologue)
        toChild.writelines(commands)
        toChild.writelines(self.cmdepilogue)
        toChild.flush()
        
        output = fromChild.readlines()
        outputError = fromChildError.readlines()
        return (output, outputError)

    



    
