import gtk

import util
from editor import Editor, EditorSyncError

# this is a generic object that provides a UI to allow the user to
# edit the parameters to an invocation and then perform that invocation by
# pressing a button.
# it also provides a "reset" button to allow the editor to be reset
# back to its initial state.
class InvocationEditor(Editor):
    # create an Invocation Editor
    #
    # name is a textual name for the invocation, like "Open"
    # desc is a description like "Open File"
    # impl is a function that impelements the operation like openFile
    # params is a sequence of tuples:
    #    (name, widget, changedHook, getValue, setValue), one
    #    for each parameter to be passed to the invocation, where
    #    name is a textual name for the parameter (like "Working directory")
    #    widget is a widget that allows the user to edit the value for
    #       this parameter (like a gtk.Entry)
    #    changedHook is a pair (w, sig) where w is a widget and sig
    #       is a signal emitted by w that indicates that the value
    #       of the user's input has changed. if None is provided,
    #       we use (widget, "changed") as default values
    #    getValue is the function to call to get the user's value,
    #       to which 'widget' will be passed
    #    setValue is the function to call to set the user's value
    #       which is used to revert the widget to its unedited state,
    #       this function is invoked with the widget and the value to set,
    #       as parameters.
    # reset is a boolean indicating whether, after each invocation,
    #         the values of each subwidget should be reset to their
    #         originals (reset==True), or kept at the current value.
    #         the default is False
    # orig  is a tuple of initial values for each parameter
    #       which defaults to None. if orig == None, or if orig[i] == None,
    #       then we take the initial value of parameter i to be the current
    #       value from the ith widget
    #
    # suppose we have two parameters,
    # i.e. params = ((n1,w1,ch1,gv1,sv1), (n2,w2,ch2,gv2,sv2)),
    # then when the button is pressed, we call impl(gv1(v1), gv2(v2))
    def __init__(self, name, desc, impl, params, reset = False, orig = None):
        Editor.__init__(self)
        (frame, vBox) = util.makeVBoxInFrame(desc)
        table = gtk.Table(len(params)+1, 2, False)
        vBox.pack_start(table, True, True, 0)
        self.initialValues = []
        startsModified = False
        labels = ()
        row = 0
        for p in params:
            (pname, widget, changedHook, getValue, setValue) = p
            
            leftattach = 0
            if pname != None:
                label = gtk.Label(pname)
                # right-align the labels by using an hbox
                labelHBox = gtk.HBox()
                labelHBox.pack_end(label, False, False, 0)
                table.attach(labelHBox,0,1,row,row+1, gtk.FILL, gtk.FILL)
                labels += ( labelHBox, )
                leftattach = 1
            else:
                labels += ( None, )
                
            table.attach(widget,leftattach,2,row,row+1)
            row += 1

            # hook the changed signal of each widget
            # pass the index of this widget to the callback
            # so we know which one changed
            
            w = widget
            sig = "changed"
            if changedHook != None:
                (w, sig) = changedHook
                
            w.connect(sig, self.subWidgetChanged,
                           len(self.initialValues))

            if type(w) == gtk.Entry:
                w.connect("activate", self.entryWidgetActivated)
                
            # get the initial values of the widgets so we can track
            # when they've been modified
            origValue = None
            widgetValue = getValue(widget)
            if orig != None:
                origValue = orig[len(self.initialValues)]

            if origValue == None:
                self.initialValues +=  [ widgetValue ]
            else:
                self.initialValues +=  [ origValue ]
                if origValue != widgetValue and not self.getModified():
                    startsModified = True
            

        button = gtk.Button(name)
        # button starts out desensitive since we're not modified
        button.set_sensitive(False)
        button.connect("clicked", self.buttonClicked, None)
        revertButton = gtk.Button("_Reset")
        revertButton.set_sensitive(False)
        revertButton.connect("clicked", self.revertButtonClicked, None)
        buttonHBox = gtk.HBox()
        buttonHBox.set_spacing(util.HSPACING)
        buttonHBox.pack_end(button, False, False)
        buttonHBox.pack_end(revertButton, False, False)
        table.attach(buttonHBox,1,2,row,row+1, gtk.FILL, gtk.FILL)
        table.set_row_spacings(util.VSPACING)
        table.set_col_spacings(util.HSPACING)

        self.table = table
        self.button = button
        self.revertButton = revertButton
        self.widget = frame
        self.impl = impl
        self.params = params
        self.reset = reset
        self.labels = labels
        
        # if we've started out in a modified state, then let everyone know
        if startsModified:
            self.modified()
            
    # get the widget
    def getWidget(self):
        return self.widget

    # this is called to indicate that we are now 'modified'
    def modified(self):
        # when we go into the modified state, ensure that both buttons
        # are sensitive
        self.button.set_sensitive(True)
        self.revertButton.set_sensitive(True)
        # pass the message on
        Editor.modified(self)

    # this is called to indicate that we are now 'synced'
    def synced(self):
        # when we go into the synced state, ensure that both buttons
        # are insensitive
        self.button.set_sensitive(False)
        self.revertButton.set_sensitive(False)
        # pass the message on
        Editor.synced(self)
    
        
    # a callback for the button press
    def buttonClicked(self, button, param):
        assert param == None
        assert button == self.button
        self.doInvoc()

    # an entry subwidget has been activated -- press the button
    def entryWidgetActivated(self, widget):
        assert type(widget) == gtk.Entry
        if self.button.get_property("sensitive"):
            self.button.clicked()
        
    # returns True if successful, False otherwise
    def doInvoc(self):
        values = self.getCurrentValues()
        if self.impl(*values):

            if (self.reset):
                index = 0
                for p in self.params:
                    (name, widget, changedHook, getValue, setValue) = p
                    setValue(widget,self.initialValues[index])
                    index += 1
            else:
                # update the initial values
                self.initialValues = values
            
            self.synced()
            return True
        return False

    def commit(self):
        if self.doInvoc():
            assert not self.getModified()
            return
        else:
            raise EditorSyncError("Error performing invocation to commit editor.")

    def revert(self):
        index = 0
        for p in self.params:
            (name, widget, changedHook, getValue, setValue) = p
            setValue(widget,self.initialValues[index])
            index += 1
        self.synced()
        
        
    # get the current values of each of the subwidgets
    def getCurrentValues(self):
        values = []
        for p in self.params:
            (name, widget, changedHook, getValue, setValue) = p            
            values += [ getValue(widget) ]
        return values

    # callback for when the reset button is clicked to reset the
    # widgets to their initial values
    def revertButtonClicked(self, revertButton, param):
        assert param == None
        self.revert()

    
    # a callback for when a subwidget's value changes
    def subWidgetChanged(self, widget, index):
        (name, wwidget, changedHook, getValue, setValue) = self.params[index]
        value = getValue(wwidget)
        initialValue = self.initialValues[index]
        if value != initialValue:
            self.modified()
        else:
            # check all widgets values. if they match their initial values
            # we're unmodified
            values = self.getCurrentValues()
            if values == self.initialValues:
                self.synced()

    # detatch a subwidget from the internal gtk.Table in which widgets
    # are packed. is useful when the same widget is shared amongst
    # multiple InvocationEditors
    def detatchWidgetForParam(self, index):
        (pname, widget, changedHook, getValue, setValue) = self.params[index]
        self.table.remove(widget)

    # reattach a previously detatched subwidget. is useful when the same
    # widget is shared amongst multiple InvocationEditors
    def reattatchWidgetForParam(self, index):
        (pname, widget, changedHook, getValue, setValue) = self.params[index]
        self.table.attach(widget,1,2,index,index+1)        

    def detatchLabelForParam(self, index):
        label = self.labels[index]
        self.table.remove(label)

    def reattatchLabelForParam(self, index):
        label = self.labels[index]
        assert label != None
        self.table.attach(label,0,1,index,index+1, gtk.FILL, gtk.FILL)
        
    # this method allows us to tweak what the initial value of a param
    # is -- it's a bit of a hack to allow changes in the underlying
    # thing that the editor is editing to be reflected in the editor,
    # even though the changes weren't made to happen by /this/ particular
    # editor. It's useful when multiple editors are used to update the same
    # underlying thing.
    def modifyInitialValueOfParam(self, index, newInitialValue):
        self.initialValues[index] = newInitialValue
        (pname, widget, changedHook, getValue, setValue) = self.params[index]
        value = getValue(widget)
        if value != newInitialValue and not self.getModified():
            self.modified()
        else:
            # check all widgets values. if they match their initial values
            # we're unmodified
            values = self.getCurrentValues()
            if values == self.initialValues:
                self.synced()
            
