from graph import Graph
from fdr import (errorDuringNewSession, errorDuringLoad, errorDuringCompile,
                 FDRCommandError)
import os
import time

# Building an ISM can be quite time consuming at the moment.
# We use the 'callbacks' object to keep the client appraised of the
# situation.
#
# callbacks should be an object supporting the following methods:
#    makeISMLog(self, msg)
#       used to send log messages
#
def makeISM(fdrObj, cspFileName, process, linesToSections, callbacks = None,
            fdrCallbacks = None):

    def wasErrOutput(out, err):
        if err != "":
            return True
        return False

    def noStdOutput(out, err):
        if out == "":
            return True
        return False

        
    def runCmd(instance, command, errCheck):
        (out, err) = instance.runCmd(command, fdrCallbacks)
        if errCheck(out,err):
            raise FDRCommandError("Error running FDR command: %s\n" % command, out, err, linesToSections)
        return (out, err)


    instance = fdrObj.newInstance()
    runCmd(instance, "session mysession", errorDuringNewSession)
    (dir, fname) = os.path.split(cspFileName)
    
    runCmd(instance, "mysession load %s %s" % (dir, fname),
           errorDuringLoad)

    runCmd(instance, "mysession compile [mysession evaluate \"%s\"] myism" % process,
           errorDuringCompile)

    name = process

    (out, err) = runCmd(instance, "puts [myism root]", noStdOutput)
    root = out.split("\n")[0]

    (out, err) = runCmd(instance, "puts [myism transitions]", noStdOutput)
    transitions = out

    (out, err) = runCmd(instance, "puts [myism numeric_alphabet]", wasErrOutput)
    numericAlphabet = out

    def log(msg):
        callbacks.makeISMLog(msg)
    
    log("Splitting numericAlphabet...\n")
    numericAlphabet = numericAlphabet.split()
    log("Getting event names...\n")
    
    eventNames = {0: "_tau", 1: "_tick"}
    for event in numericAlphabet:
        event = int(event)
        (out, err) = runCmd(instance, "puts [myism event %d]" % event,
                            wasErrOutput)
        eventNames[event] = out.split("\n")[0]

    instance.close()

    graph = Graph(name)
    nodeMap = {}

    root = int(root)
    nodeMap[root] = graph.addNode(str(root))

    log("Splitting transitions...\n")
    transitions = transitions.split('{')
    numTransitions = len(transitions)
    count = 0
    log("%d transitions. Adding each to graph...\n" % numTransitions)
    lastTime = time.time()
    for t in transitions:
	trans = t.split()
        if t.lstrip() != "":
            srcNode = int(trans[0])
            dstNode = int(trans[2].split('}')[0])
            event = int(trans[1])
            eventName =  eventNames[event]
            if srcNode not in nodeMap.keys():
                nodeMap[srcNode] = graph.addNode(str(srcNode))
            if dstNode not in nodeMap.keys():
                nodeMap[dstNode] = graph.addNode(str(dstNode))
            graph.addEdge(nodeMap[srcNode], eventName, nodeMap[dstNode])
        count += 1
        newTime = time.time()
        if newTime - lastTime > 5:
            ratio = float(count) / float(numTransitions)
            ratio = ratio * 100
            log("   Progress: %f %%\n" % ratio)
            lastTime = newTime
        
    log("   Done.\n")
    root = nodeMap[int(root)]
    return (graph, root)

    
    
    
