import threading
import gtk
import gobject
import util
import logwindow
import traceback

def refString(refinement):
    return refinement[0] + " [= " + refinement[1]

STOCK_TESTING = gtk.STOCK_EXECUTE
STOCK_PASSED = gtk.STOCK_YES
STOCK_FAILED = gtk.STOCK_NO

    
class RefinementChecker(threading.Thread):

    # refinements is a dictionary whose keys are pairs, (a, b),
    # "a refines b", and whose values are also pairs (descr, cexviewer)
    # where descr is a string describing the refinement test
    # and cexviewer is a string of Python code that defines the class
    # CounterExampleViewer (conforming to the API in counterexampleviewer.py)
    # if cexviewer is None, we use the default
    # CounterExampleViewer class from counterexampleviewer.py
    def __init__(self, name, desc, fdr, refinements, cspFilename,
                 linesToSections):
        self.fdr = fdr
        self.refinements = refinements
        self.cspFilename = cspFilename
        self.name = name
        self.desc = desc
        self.linesToSections = linesToSections
        
        self.liststore = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_STRING, object)


        for refinement in refinements.keys():
            (descr, cexviewer) = refinements[refinement]
            self.liststore.append((None, descr, refinement))

        self.treeview = gtk.TreeView(self.liststore)
        cell = gtk.CellRendererPixbuf()
        column = gtk.TreeViewColumn("Status", cell, stock_id=0)
        self.treeview.append_column(column)        
        cell = gtk.CellRendererText()
        column = gtk.TreeViewColumn("Property", cell, text=1)
        self.treeview.append_column(column)

        refCell = gtk.CellRendererText()
        column = gtk.TreeViewColumn("Refinement Test", refCell)

        def refCellDataFunc(treeviewcolumn, cell, model, iter):
            assert cell == refCell
            assert treeviewcolumn == column
            assert model == self.liststore
            
            refinement = model.get_value(iter, 2)
            assert len(refinement) == 2
            cell.set_property("text", refString(refinement))
        
        column.set_cell_data_func(refCell, refCellDataFunc)
        
        self.treeview.append_column(column)
        self.treeview.get_selection().connect("changed", self.selectChanged)
        
        (self.frame, vbox) = util.makeVBoxInFrame(name, desc)
        self.vpane = gtk.VPaned()
        vbox.pack_start(self.vpane)

        vbox = gtk.VBox()
        vbox.set_spacing(util.VSPACING)
        self.vpane.pack1(vbox, True, True)
        self.hpane = gtk.HPaned()
        
        scrolledwindow = gtk.ScrolledWindow()
        scrolledwindow.set_policy(gtk.POLICY_AUTOMATIC,
                                  gtk.POLICY_AUTOMATIC)
        scrolledwindow.add(self.treeview)
        self.hpane.pack1(scrolledwindow, True, False)
        
        vbox.pack_start(self.hpane, True, True)
        self.progressbar = gtk.ProgressBar()
        vbox.pack_start(self.progressbar, False, False)
        
        self.logWindow = logwindow.LogWindow("Log for %s" % self.name)
        self.vpane.pack2(self.logWindow.getWidget(), False, True)
        self.assertions = self.refinements.keys()
        self.errOutputTag = self.logWindow.textbuffer.create_tag(None, foreground="dark gray", foreground_set=True)

        # maps from keys in self.refinements to None if passed or
        # FDRCounterExample if failed
        self.results = {}

        
        self.cexContainer = gtk.Viewport()
        scrolledwindow = gtk.ScrolledWindow()
        scrolledwindow.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scrolledwindow.add(self.cexContainer)
        
        self.hpane.add2(scrolledwindow)

        self.noneSelectedLabel = gtk.Label("Select a refinement test from the list opposite to see its status.")
        self.noneSelectedLabel.set_line_wrap(True)

        self.noResultsLabel = gtk.Label("No results yet for this refinement.")
        self.succeededLabel = gtk.Label("Test succeeded.")

        self.cexContents = None
        
        threading.Thread.__init__(self)

    def selectChanged(self, selection):
        assert selection == self.treeview.get_selection()

        (model, iter) = selection.get_selected()
        assert model == self.liststore

        if self.cexContents != None:
            self.cexContainer.remove(self.cexContents)
            self.cexContents = None
            
        if iter != None:
                
                
            refinement = self.liststore.get_value(iter, 2)

            if refinement in self.results:
                self.cexContents = self.results[refinement]
            else:
                self.cexContents = self.noResultsLabel
                
        else:
            self.cexContents = self.noneSelectedLabel

        assert self.cexContents != None
        self.cexContainer.add(self.cexContents)
        self.cexContainer.show()
        self.cexContents.show()
            
            
    def getWidget(self):
        return self.frame
    
    def run(self):
        def do():
            res = self.fdr.checkRefinements(self.cspFilename,
                                            self.assertions,
                                            self, 1,
                                            self.fdr.VERBOSITY_MEDIUM,
                                            self.linesToSections)

            
            if res != (None,) * len(self.assertions):
                raise Exception("At least one refinement doesn't hold. \nSelect failed refinements above to view their counterexamples.\n")

        self.logWindow.performAndLog(do, (), "Checking all refinements", False, True)

    def stop(self):
        print "stop called on %s" % str(self)
        # FIXME: need some way to signal the running thread to stop

    # callback from fdr.checkRefinements() telling us that a new assertion
    # is being checked
    def testingAssertion(self, index):
        assert index < len(self.assertions)
        total = len(self.assertions)
        text = "Checking refinement %d of %d" % (index + 1, total)
        gtk.threads_enter()
        self.progressbar.set_text(text)
        iter = self.liststore.get_iter((index,))
        self.liststore.set_value(iter, 0, STOCK_TESTING)
        self.treeview.get_selection().select_iter(iter)
        gtk.threads_leave()
        
    # callback from fdr.checkRefinements() telling us that an assertion
    # has finished being checked
    def testedAssertion(self, index, res):
        assert index < len(self.assertions)

        total = len(self.assertions)
        progress = float(index+1) / float(total)

        gtk.threads_enter()
        
        # populate the results structure
        ass = self.assertions[index]
        if res != None:
            cexviewer = None
            # grab the cexviewer for this refinement which is a string of
            # python code that should define the class CounterExampleViewer
            cexMake = self.refinements[ass][1]
            if cexMake != None:
                # exec the code, thereby defining the class
                # this only affects the current scope. the definition
                # is lost once we leave the current function which is
                # what we want
                try:
                    exec cexMake
                except Exception, e:
                    cexMake = None
            if cexMake == None:
                # fallback to default
                from counterexampleviewer import CounterExampleViewer

            # build the viewer
            try:
                cexviewer = CounterExampleViewer(res)
            except Exception, e:
                cexviewer = None
            if cexviewer == None:
                from counterexampleviewer import CounterExampleViewer
                try:
                    cexviewer = CounterExampleViewer(res)
                except Exception, e:
                    cexviewer = None
            assert cexviewer != None
            
            self.results[ass] = cexviewer.getWidget()

        else:
            self.results[ass] = self.succeededLabel
        
        self.progressbar.set_fraction(progress)
        if index + 1 == total:
            self.progressbar.set_text("Complete")
        self.liststore.set_value(self.liststore.get_iter((index,)), 0, (STOCK_FAILED, STOCK_PASSED)[res == None])
        self.treeview.get_selection().unselect_all()
        gtk.threads_leave()
        

    # callback from fdr.checkRefinements() indicating new stdoutput from FDR
    def newFDROutput(self, text):
        # we ignore stdoutput from FDR for now since it doesn't
        # usually contain anything of interest and contains too much
        # garbage like output to "puts [ cmd ]" commands
        pass

    # callback from fdr.checkRefinements() indicating new stderr outut from FDR
    def newFDRErrOutput(self, text):
        gtk.threads_enter()
        self.logWindow.log(text, False, self.errOutputTag)
        gtk.threads_leave()

        
            
        
