import util
import expr
from behaviour import Behaviour
import xml.dom.minidom
import re
from editor import Editor
from authority import Authority

OPTION_PARAMETERISEDOPS = "parameterisedops"

XML_FORMAT_VERSION = "0.2"

nameRegexp = r"^\w+'*$"
namePattern = re.compile(nameRegexp)
    
    
channelDeclRegexp = r"^\s*channel"
channelDeclPattern = re.compile(channelDeclRegexp, re.MULTILINE)
# check whether the given code is valid
# throws an exception if invalid code found
def validateCode(code):
    if channelDeclPattern.search(code) != None:
        raise ValueError("Invalid code: channel declarations prohibited")

def validateAllowedAuthority(authority):
    if authority.perms == None and authority.wpc == None and authority.wnc == None and authority.spc == None and authority.snc == None:
        raise ValueError("Invalid allowed authority: all values cannot be None")

# represents a system
# a System is an Editor for the underlying XML file that the system was
# (last) read from. Hence why it sublcasses Editor
class System(Editor):

    def __init__(self, xmlFile = None):
        # default values
        self.ops = []
        self.parameterisedops = False
        self.objects = {}    # maps object names to Behaviour objects
        self.behaviours = {} # maps abstract behaviour names to code
        self.refines = {}    # maps abstract behaviour names to behaviours
        self.code = ""
        self.allowedAuthority = {} # maps object names to Authority objects
        self.assertions = {} # maps assertion names to (name, args) pairs
        
        Editor.__init__(self)
        
        # tuple of concrete behaviour refinement asserts derived from the
        # concrete behaviours instantiated by the objects in the system
        # and the refinement relationships that exist between the
        # abstract behaviours that the concrete behaviours instantiate
        #
        # this has the form
        # ((a1, a2), (a3, a4) ...) meaning that concrete process a1 is
        # refined by concrete process a2, and concrete process a3 is refined by
        # concrete process a4 etc.
        self.concreteBehaviourAsserts = ()
        
        if xmlFile != None:
            # parse XML
            self.makeFromXML(xmlFile)
            

    def makeFromXML(self, xmlFile):

        
        def procOpNode(self, node):
            self.addOperation(node.attributes["name"].value)
            
        def procOptionNode(self, node):
            def toBoolean(value):
                if value.upper() == "FALSE":
                    value = False
                else:
                    value = True
                return value
            
            name = node.attributes["name"].value
            value = node.attributes["value"].value

            if name == OPTION_PARAMETERISEDOPS:
                value = toBoolean(value)
                self.setParameterisedOps(value)

        def procAssertionNode(self, node):
            name = node.attributes["name"].value
            head = node.attributes["head"].value
            args = ()
            for child in node.childNodes:
                if (child.nodeType == child.ELEMENT_NODE and
                    child.localName == "arg"):
                    arg = child.attributes["value"].value
                    args += (arg,)
            # args == () doesn't make sense; it confuses the systemeditor
            if args == ():
                args = None
            self.addAssertion(name, (head, args))

        # returns a Behaviour object
        def procBehaviourNode(node):
            name = node.attributes["name"].value
            args = ()
            for child in node.childNodes:
                if (child.nodeType == child.ELEMENT_NODE and
                    child.localName == "arg"):
                    arg = child.attributes["value"].value
                    args += (arg,)
            # args == () doesn't make sense; it confuses the systemeditor
            if args == ():
                args = None
            return Behaviour(name, args)
        
            
        def procObjectNode(self, node):

            def procAuthorityNode(node):

                def getCDataChild(node):
                    for child in node.childNodes:
                        if (child.nodeType == node.CDATA_SECTION_NODE):
                            return child.data
                    raise ValueError("CDATA section node not found")

                perms = None
                wpc = None
                wnc = None
                spc = None
                snc = None
                dpc = None
                dnc = None
                
                got = False
                for child in node.childNodes:
                    if child.nodeType == child.ELEMENT_NODE:
                        if child.localName == "perms":
                            perms = getCDataChild(child)
                            got = True
                        elif child.localName == "wpc":
                            wpc = getCDataChild(child)
                            got = True
                        elif child.localName == "wnc":
                            wnc = getCDataChild(child)
                            got = True
                        elif child.localName == "spc":
                            spc = getCDataChild(child)
                            got = True
                        elif child.localName == "snc":
                            snc = getCDataChild(child)
                            got = True
                        elif child.localName == "dpc":
                            dpc = getCDataChild(child)
                            got = True
                        elif child.localName == "dnc":
                            dnc = getCDataChild(child)
                            got = True
                if not got:
                    raise ValueError("empty authority node")
                authority = Authority(perms, wpc, wnc, spc, snc, dpc, dnc)
                return authority
            
                            
            name = node.attributes["name"].value
            behaviour = None
            authority = None
            
            for child in node.childNodes:
                if child.nodeType == child.ELEMENT_NODE:
                    if child.localName == "behaviour":
                        # found behaviour node
                        behaviour = procBehaviourNode(child)
                    elif child.localName == "allowedauthority":
                        for cchild in child.childNodes:
                            if (cchild.nodeType == child.ELEMENT_NODE and
                                cchild.localName == "authority"):
                                authority = procAuthorityNode(cchild)
                        if authority == None:
                            raise ValueError("No authority node for allowedauthority node of object %s" % name)
                        
            if behaviour == None:
                raise ValueError("No behaviour node for object %s" % name)
                                    
            self.addObject(name,behaviour)
            if authority != None:
                self.updateAllowedAuthority(name, authority)

        def procAbstractbehaviourNode(self, node):
            name = node.attributes["name"].value
            self.validateAbsBehaviourName(name)
            
            refines = None
                
            code = ""
            foundCode = False
            # get the code, if present
            for child in node.childNodes:
                if child.nodeType == child.ELEMENT_NODE:
                    if child.localName == "csp":
                        foundCode = True
                        for codeChild in child.childNodes:
                            if (codeChild.nodeType == codeChild.CDATA_SECTION_NODE):
                                code += codeChild.data
                    elif child.localName == "refines":
                        refines = procBehaviourNode(child)
                        
            if not foundCode:
                raise ValueError("No code specified for behaviour %s" % name)

            self.addBehaviour(name, code)
            if refines != None:
                self.updateRefines(name, refines)

        def procCspNode(self, node):
            for child in node.childNodes:
                if (child.nodeType == child.CDATA_SECTION_NODE):
                    code = child.data
                    validateCode(code)
                    self.code += child.data

        # parse the XML    
        xmldoc = xml.dom.minidom.parse(xmlFile).documentElement
        if (xmldoc.nodeType == xmldoc.ELEMENT_NODE and
            xmldoc.localName == "system"):

            # check for correct version
            try:
                version = xmldoc.attributes["formatVersion"].value
            except Exception, e:
                raise ValueError("No format version specified.")
            
            if version != XML_FORMAT_VERSION:
                raise ValueError("Incompatible format, expected verion %s, got %s" % (XML_FORMAT_VERSION,version))
            
            for child in xmldoc.childNodes:
                if child.nodeType == child.ELEMENT_NODE:
                    if child.localName == "op":
                        procOpNode(self, child)
                    elif child.localName == "option":
                        procOptionNode(self, child)
                    elif child.localName == "object":
                        procObjectNode(self, child)
                    elif child.localName == "abstractbehaviour":
                        procAbstractbehaviourNode(self, child)
                    elif child.localName == "assertion":
                        procAssertionNode(self, child)
                    elif child.localName == "csp":
                        procCspNode(self, child)
                    else:
                        raise ValueError("Unexpected node " + child.localName)
        else:
            raise ValueError("Expected root document node of system, not " + xmldoc.localName)
        self.synced()
        
    def __str__(self):
        s = "System" + "\n"
        s += "Operations: "
        for op in self.ops:
            s += op + ", "
        s += "\n"
        for obj in self.objects.keys():
            s += "Object " + obj + ": " + str(self.objects[obj]) + "\n"
        for b in self.behaviours.keys():
            s += "Behaviour " + b + "; code: " + self.behaviours[b] + "\n"
        for n in self.refines.keys():
            s += n + " refines " + str(self.refines[n]) + "\n"
        return s

    # write out the XML representation of the system
    def toXML(self, to):

        doc = xml.dom.minidom.Document()
        def makeAuthorityElement(authority):
            aElement = doc.createElement("authority")
            if authority.perms != None:
                perms = doc.createElement("perms")
                cdataNode = doc.createCDATASection(authority.perms)
                perms.appendChild(cdataNode)
                aElement.appendChild(perms)
            if authority.wpc != None:
                wpc = doc.createElement("wpc")
                cdataNode = doc.createCDATASection(authority.wpc)
                wpc.appendChild(cdataNode)
                aElement.appendChild(wpc)
            if authority.wnc != None:
                wnc = doc.createElement("wnc")
                cdataNode = doc.createCDATASection(authority.wnc)
                wnc.appendChild(cdataNode)
                aElement.appendChild(wnc)
            if authority.spc != None:
                spc = doc.createElement("spc")
                cdataNode = doc.createCDATASection(authority.spc)
                spc.appendChild(cdataNode)
                aElement.appendChild(spc)
            if authority.snc != None:
                snc = doc.createElement("snc")
                cdataNode = doc.createCDATASection(authority.snc)
                snc.appendChild(cdataNode)
                aElement.appendChild(snc)
            if authority.dpc != None:
                dpc = doc.createElement("dpc")
                cdataNode = doc.createCDATASection(authority.dpc)
                dpc.appendChild(cdataNode)
                aElement.appendChild(dpc)
            if authority.dnc != None:
                dnc = doc.createElement("dnc")
                cdataNode = doc.createCDATASection(authority.dnc)
                dnc.appendChild(cdataNode)
                aElement.appendChild(dnc)
                
            return aElement
        docElement = doc.createElement("system")
        docElement.setAttribute("formatVersion",XML_FORMAT_VERSION)
        doc.appendChild(docElement)

        optionElement = doc.createElement("option")
        optionElement.setAttribute("name", OPTION_PARAMETERISEDOPS)
        optionElement.setAttribute("value", ("true", "false")[self.parameterisedops == False])
        docElement.appendChild(optionElement)
        
        for op in sorted(self.ops):
            opElement = doc.createElement("op")
            opElement.setAttribute("name", op)
            docElement.appendChild(opElement)

        def populateBehaviourNode(node, behaviour):
            node.setAttribute("name", behaviour.name)
            if behaviour.args is not None:
                for arg in behaviour.args:
                    argElement = doc.createElement("arg")
                    argElement.setAttribute("value", arg)
                    node.appendChild(argElement)
            
        for obj in sorted(self.objects.keys()):
            objElement = doc.createElement("object")
            objElement.setAttribute("name", obj)
            behaviour = self.objects[obj]
            bElement = doc.createElement("behaviour")

            populateBehaviourNode(bElement, behaviour)
            
            objElement.appendChild(bElement)
            if obj in self.allowedAuthority.keys():
                aaElement = doc.createElement("allowedauthority")
                aElement = makeAuthorityElement(self.allowedAuthority[obj])
                aaElement.appendChild(aElement)
                objElement.appendChild(aaElement)
                
            docElement.appendChild(objElement)

        for absBehaviour in sorted(self.behaviours.keys()):
            absElement = doc.createElement("abstractbehaviour")
            absElement.setAttribute("name", absBehaviour)
            if absBehaviour in self.refines.keys():
                refinesNode = doc.createElement("refines")
                populateBehaviourNode(refinesNode, self.refines[absBehaviour])
                absElement.appendChild(refinesNode)
                
            cspNode = doc.createElement("csp")
            cdataNode = doc.createCDATASection(self.behaviours[absBehaviour])
            cspNode.appendChild(cdataNode)
            absElement.appendChild(cspNode)
            docElement.appendChild(absElement)

        for ass in sorted(self.assertions.keys()):
            assertionElement = doc.createElement("assertion")
            assertionElement.setAttribute("name", ass)
            assertion = self.assertions[ass]
            assertionElement.setAttribute("head", assertion[0])
            if assertion[1] is not None:
                for arg in assertion[1]:
                    argElement = doc.createElement("arg")
                    argElement.setAttribute("value", arg)
                    assertionElement.appendChild(argElement)
            docElement.appendChild(assertionElement)

        cspNode = doc.createElement("csp")
        cdataNode = doc.createCDATASection(self.code)
        cspNode.appendChild(cdataNode)
        docElement.appendChild(cspNode)
            
        to.write(doc.toprettyxml("   ","\n","UTF-8"))
        self.synced()
        
    # write out a CSP file that represents a system
    # returns a dictionary on success that maps from line numbers in the
    # file to different strings that name different parts of the file
    # to allow users to identify where an error in the CSP code pertains
    # to in their systems.
    def exportToCSP(self, to):
        global output
        global lastSection
        global lastStart

        if len(self.ops) == 0:
            raise Exception("Cannot export system with no Operations to CSP.")

        if len(self.objects) == 0:
            raise Exception("Cannot export system with no Objects to CSP.")
        
        lastSection = None
        lastStart = None
        output = ""
        
        # map line numbers in the exported CSP to the section of the system
        # they pertain to
        linesToSection = {}
        
        def assString(ass):
            return "assert %s [FD= %s\n" % ass

        # begin a new section of output
        def beginSection(name):            
            global lastSection
            global lastStart
            global output
            
            assert lastSection == None
            assert lastStart == None
            
            currLines = len(output.split("\n"))
            assert currLines not in linesToSection
            linesToSection[currLines] = name
            lastSection = name
            lastStart = currLines
            
            # separate all sections by some newlines
            output += "-- begin section %s\n" % lastSection
            

        # end the current section of output
        def endSection():
            global lastSection
            global lastStart
            global output
            
            assert lastSection != None
            assert lastStart != None
            
            assert linesToSection[lastStart] == lastSection

            # separate all sections by some newlines
            output += "-- end section %s \n" % lastSection
            
            endLine = len(output.split("\n")) - 1
            line = lastStart + 1
            while line <= endLine:
                try:
                    assert line not in linesToSection
                except Exception, e:
                    raise
                
                linesToSection[line] = lastSection
                line += 1
            lastSection = None
            lastStart = None
            
            
        output += ("-- CSP code for system\n")
        beginSection("Operations")
        output += ("datatype Operation = ")
        output += (util.seqString(self.ops," | ") + "\n")
        endSection()

        beginSection("Objects")
        output += ("datatype Object = ")
        output += (util.seqString(self.objects.keys()," | ") + "\n")
        endSection()
        
        if self.parameterisedops:
            output += ("datatype Data = null\n")
            output += ("Params = union({ x | x <- Object},Data)\n")
            output += ("channel send, recv : Object.Operation.Params\n")
            output += ("channel ioe_ : Object.Object.Operation.Params\n")
            output += ("toOpTransform_(P,obj) = P[[recv.o.op.p <- ioe_.o.obj.op.p, send.o.op.p <- ioe_.obj.o.op.p | o <-diff(Object,{obj}), op <- Operation, p <- Params]]\n")            
        else:
            output += ("channel send,recv : Object.Operation\n")
            output += ("channel ioe_ : Object.Object.Operation\n")
            output += ("toOpTransform_(P,obj) = P[[recv.o.op <- ioe_.o.obj.op, send.o.op <- ioe_.obj.o.op | o <-diff(Object,{obj}), op <- Operation]]\n")

        output += ("""alphabet(p) = Union({ {| ioe_.p.o | o <- diff(Object,{p}) |},
                                  {| ioe_.o.p | o <- diff(Object,{p})|} })\n""")
        output += ("theAlphabet = {|ioe_|}\n")

        beginSection("Ancillary Code")
        output += (self.code + "\n")
        endSection()
        
        for be in self.behaviours.keys():
            beginSection("Behaviour %s" % be)
            output += (self.behaviours[be] + "\n")
            endSection()

        for obj in self.objects.keys():
            beginSection("Object %s" % obj)
            # ensure that the object is prevented from trying to
            # recv from or send to itself to prevent spurious events
            # showing up in the object's individual behaviour. This doesn't
            # effect the final composition of the System since the
            # alphabetised parallel composition used to form System prevents
            # the occurrence of these (renamed) events anyway
            output += ("behaviour(%s) = %s [| diff(Events,{|send.o, recv.o | o <- diff(Object,{%s})|}) |] STOP\n" % (obj, str(self.objects[obj]), obj))
            endSection()
            
        output += ("\n")
        
        output += ("System = || i:Object @ [alphabet(i)] toOpTransform_(behaviour(i),i)\n")
        output += ("\n")
        
        for ass in self.concreteBehaviourAsserts:
            output += ("%s" % assString(ass))

        output += ("renameLocalEvents_(S,obj) = { renameLocalEvents_'(s,obj) | s <- S}\n")
        if self.parameterisedops:
            output += ("renameLocalEvents_'(send.other.op.p,obj) = ioe_.obj.other.op.p\n")
            
            output += ("renameLocalEvents_'(recv.from.op.p,obj) = ioe_.from.obj.op.p\n")        
        else:
            output += ("renameLocalEvents_'(send.other.op,obj) = ioe_.obj.other.op\n")
            
            output += ("renameLocalEvents_'(recv.from.op,obj) = ioe_.from.obj.op\n")        
        output += ("renameLocalEvents_'(other,obj) = other\n")
                                             
        for obj in self.allowedAuthority.keys():
            authority = self.allowedAuthority[obj]
            if authority.perms != None:
                beginSection("Allowed Acquirable Permissions of %s" % obj)
                output += ("allowedPerms_(%s) = renameLocalEvents_(%s,%s)\n" %
                         (obj, authority.perms, obj))
                endSection()
                
        for obj in self.allowedAuthority.keys():
            authority = self.allowedAuthority[obj]
            if authority.wpc != None:
                beginSection("Allowed WPC of %s" % obj)                
                output += ("allowedWPC_(%s) = renameLocalEvents_(%s,%s)\n" %
                         (obj, authority.wpc, obj))
                endSection()
                
        for obj in self.allowedAuthority.keys():
            authority = self.allowedAuthority[obj]
            if authority.wnc != None:
                beginSection("Allowed WNC of %s" % obj)
                output += ("allowedWNC_(%s) = renameLocalEvents_(%s,%s)\n" %
                         (obj, authority.wnc, obj))
                endSection()
                
        for obj in self.allowedAuthority.keys():
            authority = self.allowedAuthority[obj]
            if authority.spc != None:
                beginSection("Allowed SPC of %s" % obj)                
                output += ("allowedSPC_(%s) = renameLocalEvents_(%s,%s)\n" %
                         (obj, authority.spc, obj))
                endSection()
                
        for obj in self.allowedAuthority.keys():
            authority = self.allowedAuthority[obj]
            if authority.snc != None:
                beginSection("Allowed SNC of %s" % obj)                
                output += ("allowedSNC_(%s) = renameLocalEvents_(%s,%s)\n" %
                         (obj, authority.snc, obj))
                endSection()

        for obj in self.allowedAuthority.keys():
            authority = self.allowedAuthority[obj]
            if authority.dpc != None:
                beginSection("Allowed DPC of %s" % obj)                
                output += ("allowedDPC_(%s) = renameLocalEvents_(%s,%s)\n" %
                         (obj, authority.dpc, obj))
                endSection()

        for obj in self.allowedAuthority.keys():
            authority = self.allowedAuthority[obj]
            if authority.dnc != None:
                beginSection("Allowed DNC of %s" % obj)                
                output += ("allowedDNC_(%s) = renameLocalEvents_(%s,%s)\n" %
                         (obj, authority.dnc, obj))
                endSection()
                
        to.write(output)
        
        return linesToSection
    
    def validateName(self, name):
        if namePattern.match(name) == None:
            raise ValueError("Invalid name %s" % name)
        # check it doesn't clash with any of the other names in the system
        if name in self.objects.keys():
            raise ValueError("An object with the name %s already exists" % name)
        if name in self.ops:
            raise ValueError("An operation with the name %s already exists" % name)
        for absBehaviour in self.behaviours.keys():
            (absName, params) = expr.parse(absBehaviour)
            if name == absName:
                raise ValueError("A behaviour with the name %s already exists" % name)

    # check whether the given "refines" is valid for the given abstract
    # behaviour name, "name". raises an exception if not valid
    def validateAbsBehaviourRefines(self, refines, name):
        # noop for now
        return
        
    # check whether the given abstract behaviour name
    # "name" is valid. raises an exception if not valid
    def validateAbsBehaviourName(self, name):
        try:
            (nam, params) = expr.parse(name)
        except:
            raise ValueError("Invalid behaviour name %s" % name)
        self.validateName(nam)

    # add an object to the system
    def addObject(self, name, behaviour):
        self.validateName(name)        
        self.objects[name] = behaviour
        self.modified()

    # remove an object from the system
    def removeObject(self, name):
        if name not in self.objects.keys():
            raise ValueError("Object %s doesn't exist" % name)
        del self.objects[name]
        if name in self.allowedAuthority.keys():
            del self.allowedAuthority[name]
        self.modified()

    # update an object in the system
    def updateObject(self, name, behaviour):
        if name not in self.objects.keys():
            raise ValueError("Object %s doesn't exist" % name)
        self.objects[name] = behaviour
        self.modified()


    # add an operation
    def addOperation(self, name):
        self.validateName(name)
        self.ops += [ name ]
        self.modified()

    # remove an operation
    def removeOperation(self, name):
        if name not in self.ops:
            raise ValueError("Operation %s not defined" % name)
        self.ops.remove(name)
        self.modified()

    # update the code
    def updateCode(self, newCode):
        validateCode(newCode)
        self.code = newCode
        self.modified()

    # add an abstract behaviour
    def addBehaviour(self, behaviour, code):
        self.validateAbsBehaviourName(behaviour)        
        validateCode(code)
        self.behaviours[behaviour] = code
        self.modified()

    # remove an abstract behaviour
    def removeBehaviour(self, behaviour):
        if behaviour not in self.behaviours.keys():
            raise ValueError("Behaviour %s not defined" % behaviour)
        del self.behaviours[behaviour]
        if behaviour in self.refines:
            del self.refines[behaviour]
        self.modified()

    # update the code of a behaviour
    def updateBehaviour(self, behaviour, newCode):
        validateCode(newCode)
        if behaviour not in self.behaviours.keys():
            raise ValueError("Behaviour %s not defined" % behaviour)
        self.behaviours[behaviour] = newCode
        self.modified()
        
    # update the refines relationship for an abstract behaviour
    def updateRefines(self, behaviour, refines):
        if behaviour not in self.behaviours.keys():
            raise ValueError("Behaviour %s not defined" % behaviour)
        if refines != None:
            self.validateAbsBehaviourRefines(refines, behaviour)        
            self.refines[behaviour] = refines
        else:
            assert behaviour in self.refines.keys()
            del self.refines[behaviour]
        self.modified()

    # build the asserts for each of the objects in 'objects'
    # if 'objects' is none, we do all of the objects in the system
    def buildConcreteBehaviourAsserts(self, objects=None):

        # clear the current ones
        self.concreteBehaviourAsserts = ()
        
        doneRefinesFor = []

        if objects == None:
            objects = self.objects.keys()
                
        # for each object, get its behaviour and
        # then get the concrete instantiation of the behaviour it refines
        # based on how it instantiates its abstract behaviour
        # then get the behaviour that /that/ concrete instantiation refines
        # and so on
        for obj in objects:
            behaviour = self.objects[obj]
            assert behaviour != None
            
            concreteRefines = behaviour.getConcreteRefines(self)

            
            # we have to remember which behaviours we've already done, in
            # case we encounter a cycle
            # this is what doneRefinesFor is used for
            # but we need immutable objects so that equality comparisons
            # work correctly so we use strings instead of the objects
            # themselves
            behaviourStr = str(behaviour)
            concreteRefinesStr = str(concreteRefines)
            while behaviourStr not in doneRefinesFor:
                doneRefinesFor += [ behaviourStr ]
                
                if concreteRefines != None:
                    self.concreteBehaviourAsserts += (
                        (concreteRefinesStr, behaviourStr), )
                
                    behaviour = concreteRefines
                    concreteRefines = behaviour.getConcreteRefines(self)
                    behaviourStr = concreteRefinesStr
                    concreteRefinesStr = str(concreteRefines)

    # update an object's allowed authority. None means we're not interested
    # in this object's authority
    def updateAllowedAuthority(self, object, authority):
        if authority != None:
            validateAllowedAuthority(authority)
            self.allowedAuthority[object] = authority
        else:
            if object in self.allowedAuthority.keys():
                del self.allowedAuthority[object]
        self.modified()
    
    def setParameterisedOps(self, value):
        if value:
            self.parameterisedops = True
        else:
            self.parameterisedops = False
        self.modified()

    # add an assertion to the system
    def addAssertion(self, name, assertion):
        if name in self.assertions.keys():
            raise ValueError("Assertion %s already exists" % name)
        
        self.assertions[name] = assertion
        self.modified()

    # remove an assertion from the system
    def removeAssertion(self, name):
        if name not in self.assertions.keys():
            raise ValueError("Assertion %s doesn't exist" % name)
        del self.assertions[name]
        self.modified()

    # update an assertion in the system
    def updateAssertion(self, name, assertion):
        if name not in self.assertions.keys():
            raise ValueError("Assertion %s doesn't exist" % name)
        self.assertions[name] = assertion
        self.modified()

# we have to make these 'global' so our nested functions can access them
# properly -- Python or I suck here, I'm not sure which
lastSection = None
lastStart = None
output = ""
        
