import pygtk
import gtk
import gobject
import pango
if gtk.pygtk_version < (2,6,0) or gtk.gtk_version < (2,6,0):
    raise ImportError("PyGTK 2.6.0 and GTK 2.6.0 required")
import sys

import util
import expr
from behaviour import Behaviour
from editor import Editor, EditorSyncError
from invocationeditor import InvocationEditor
from sourcecodeeditor import SourceCodeEditor
from action import Action
from authority import Authority

SYSTEM_NAME = "System"

NODE_OPS = "Operations"
NODE_OBJECTS = "Objects"
NODE_ABSBEHAVIOURS = "Behaviours"
NODE_ASSERTIONS = "Assertions"
NODE_CODE = "Ancillary Code"

FQ_NAME_SEP = " / "

STATUS_SYNCED = ""
STATUS_MODIFIED = "(modified)"

    
def getEditString(name):
    return "Edit the %s for this system." % name.lower()

def getNotEditableString(name):
    return "No editable properties for %s." % name

def getTextFromEntry(entry):
    return entry.get_text()

def setTextInEntry(entry, text):
    return entry.set_text(text)

def getBoolFromCheck(check):
    return check.get_active()

def setBoolInCheck(check, value):
    check.set_active(value)


# a base class used to aggregate a number of sub Editor objects
# we are 'modified' iff any of the subedotrs are
# we are 'synced' iff all of the subeditors are
class AggregatedEditor(Editor):
    def __init__(self):
        Editor.__init__(self)
        self.editors = []

    def addEditor(self, editor):
        assert editor not in self.editors
        editor.addModifiedCallback(self.subEditorModified)
        editor.addSyncedCallback(self.subEditorSynced)
        self.editors += [editor,]
        if editor.getModified() and not self.getModified():
            self.modified()

    def removeEditor(self, editor):
        assert editor in self.editors
        editor.removeModifiedCallback(self.subEditorModified)
        editor.removeSyncedCallback(self.subEditorSynced)
        self.editors.remove(editor)
        
    def subEditorModified(self, editor):
        assert editor in self.editors
        if not self.getModified():
            self.modified()

    def subEditorSynced(self, editor):
        assert editor in self.editors
        for ed in self.editors:
            if ed.getModified():
                break
        else:
            self.synced()
    
    def commit(self):
        for ed in self.editors:
            if ed.getModified():
                ed.commit()

    def revert(self):
        for ed in self.editors:
            if ed.getModified():            
                ed.revert()
    
    def getWidget(self):
        raise NotImplementedError("AggregatedEditor.getWidget() not implemented")
    
# an editor for a system's operations    
class OperationsEditor(AggregatedEditor):
    def __init__(self, fqName, systemeditor):
        AggregatedEditor.__init__(self)
        self.fqName = fqName
        self.systemeditor = systemeditor
        
        (frame, vbox) = util.makeVBoxInFrame(self.fqName,
                                        getEditString(NODE_OPS))

        nameEntry = gtk.Entry()
        addOp = InvocationEditor("_Add", "Add An Operation",
                                 self.addOperationImpl,
                                 (("Operation: ", nameEntry, None,
                                   getTextFromEntry, setTextInEntry),),
                                 True)
        
        
        vbox.pack_start(addOp.getWidget(), False, True, 0)
        
        self.addEditor(addOp)
        self.widget = frame

    def getWidget(self):
        return self.widget
        
    def addOperationImpl(self, name):
        if self.systemeditor.performAndLog(self.systemeditor.system.addOperation, (name,),
                                           "Adding operation %s" % name):
            self.systemeditor.addNodeToSystemTreeView(
                util.seqString((SYSTEM_NAME,
                                NODE_OPS,
                                name),FQ_NAME_SEP))
            return True
        return False
        
class OperationEditor(object):
    def __init__(self, fqName, systemeditor, name):
        self.fqName = fqName
        self.systemeditor = systemeditor
        self.name = name
        (frame, vbox) = util.makeVBoxInFrame(fqName,
                                             "Edit operation %s." % name)

        delOp = Action(None, "Delete operation %s" % name,
                       self.deleteOp, gtk.STOCK_DELETE)
        vbox.pack_start(delOp.getWidget(), False, False, 0)
        self.widget = frame
        
    def deleteOp(self):
        if self.systemeditor.performAndLog(self.systemeditor.system.removeOperation, (self.name,),
                                           "Removing operation %s" % self.name):
            self.systemeditor.removeNodeFromSystemTreeView(self.fqName)
            self.systemeditor.removeEditor(self)

    def getWidget(self):
        return self.widget
        
# an editor for a system's ancillary code        
class CodeEditor(AggregatedEditor):
    def __init__(self, fqName, systemeditor):
        AggregatedEditor.__init__(self)
        self.fqName = fqName
        self.systemeditor = systemeditor

        (frame, vbox) = util.makeVBoxInFrame(fqName,
                                        getEditString(NODE_CODE))
        
        cedit = SourceCodeEditor(self.systemeditor.system.code,
                                 self.updateCodeImpl,
                                 "Update the Ancillary Code")
        vbox.pack_start(cedit.getWidget(), True, True, 0)

        self.addEditor(cedit)
        self.widget = frame
        
    def getWidget(self):
        return self.widget

    def updateCodeImpl(self, newCode):
        return self.systemeditor.performAndLog(self.systemeditor.system.updateCode,
                                               (newCode,),
                                               "Updating ancillary code")
        
# an editor for a particular abstract behaviour    
class AbsBehaviourEditor(AggregatedEditor):
    def __init__(self, fqName, systemeditor, name):
        AggregatedEditor.__init__(self)
        self.fqName = fqName
        self.systemeditor = systemeditor
        self.name = name

        self.refinesEditorWidget = None
        self.refinesEditorWidgetContainer = gtk.VBox()

        self.refinesEditor = AbsBehaviourRefinesEditor(name, systemeditor,
                                                       self.refinesEditorWidgetChanged)
        self.addEditor(self.refinesEditor)
        

        (frame, vbox) = util.makeVBoxInFrame(fqName,
                                        "Edit behaviour %s." % name)
        
        vbox.pack_start(self.refinesEditorWidgetContainer, False, False, 0)
        
        cedit = SourceCodeEditor(self.systemeditor.system.behaviours[name],
                                 self.updateBehaviourImpl,
                                 "Update the code for %s" % name)
        vbox.pack_start(cedit.getWidget(), True, True, 0)
        self.addEditor(cedit)

        delMe = Action(None, "Delete %s" % name, self.deleteMe,
                       gtk.STOCK_DELETE)
        vbox.pack_start(delMe.getWidget(), False, False, 0)
        
        self.widget = frame
        

    def refinesEditorWidgetChanged(self, newWidget):
        if self.refinesEditorWidget != None:
            self.refinesEditorWidgetContainer.remove(self.refinesEditorWidget)
        self.refinesEditorWidgetContainer.pack_start(newWidget, False, False, 0)
        self.systemeditor.redrawWindow()
        self.refinesEditorWidget = newWidget
        
    
    # callback for when the code is updated
    def updateBehaviourImpl(self, code):
        return self.systemeditor.performAndLog(
            self.systemeditor.system.updateBehaviour,
            (self.name, code),
            "Updating code for behaviour %s" % self.name)
                        
    def updateRefinesImpl(self, refines):
        if refines.lstrip() == "":
            refines = None
        return self.systemeditor.performAndLog(
            self.systemeditor.system.updateRefines,
            (self.name, refines),
            "Updating refines for behaviour %s" % self.name)

    # callback for when this abs behaviour is to be deleted
    def deleteMe(self):
        if self.systemeditor.performAndLog(self.systemeditor.system.removeBehaviour, (self.name,),
                                           "Removing behaviour %s" % self.name):
            self.systemeditor.removeNodeFromSystemTreeView(self.fqName)
            self.systemeditor.removeNodeFromAbsBehavioursListStore(self.name)
            self.systemeditor.removeEditor(self)
            
    def getWidget(self):
        return self.widget


# an editor for a system's abstract behaviours
class AbsBehavioursEditor(AggregatedEditor):
    def __init__(self, fqName, systemeditor):
        AggregatedEditor.__init__(self)
        self.fqName = fqName
        self.systemeditor = systemeditor
        
        (frame, vbox) = util.makeVBoxInFrame(self.fqName,
                                             getEditString(NODE_ABSBEHAVIOURS))

        nameEntry = gtk.Entry()
        addAB = InvocationEditor("_Add", "Add A Behaviour",
                                 self.addAbsBehaviourImpl,
                                 (("Behaviour: ", nameEntry, None,
                                   getTextFromEntry, setTextInEntry),),
                                 True)
        
        
        vbox.pack_start(addAB.getWidget(), False, True, 0)
        
        self.addEditor(addAB)
        self.widget = frame

    def getWidget(self):
        return self.widget
        
    def addAbsBehaviourImpl(self, name):
        if self.systemeditor.performAndLog(self.systemeditor.system.addBehaviour, (name, ""),
                                           "Adding behaviour %s" % name):
            self.systemeditor.addNodeToSystemTreeView(
                util.seqString((SYSTEM_NAME,
                                NODE_ABSBEHAVIOURS,
                                name),FQ_NAME_SEP))
            self.systemeditor.addNodeToAbsBehavioursListStore(name)
            return True
        return False

# an editor for a system's system-wide properties
class SystemPropertiesEditor(AggregatedEditor):
    def __init__(self, fqName, systemeditor):
        AggregatedEditor.__init__(self)
        self.fqName = fqName
        self.systemeditor = systemeditor
        
        (frame, vbox) = util.makeVBoxInFrame(self.fqName,
                                             "Edit the system properties.")

        
        popsCheck = gtk.CheckButton("Operations are _Parameterised")
        popsCheck.set_active(systemeditor.system.parameterisedops)
        setpops = InvocationEditor("_Commit", "Parameterised Operations",
                                 self.setpopsImpl,
                                   ((None,
                                     popsCheck, (popsCheck, "toggled"),
                                     getBoolFromCheck,
                                     setBoolInCheck),),
                                   False)
        
        
        vbox.pack_start(setpops.getWidget(), False, True, 0)
        
        self.addEditor(setpops)
        self.widget = frame

    def getWidget(self):
        return self.widget
        
    def setpopsImpl(self, value):
        return self.systemeditor.performAndLog(self.systemeditor.system.setParameterisedOps, (value,),
                                           "Setting 'parameterised ops' option to %s" % str(value))

# editor for a concrete Expr -- like an object's behaviour or an
# instantiated predicate.
# we build a different Editor for each abstract expr on demand.
# all of these Editors share the same combobox
# but only one of them (corredponding to the currently selected entry
# in the combobox) is ever registered as a sub-editor at a time.
# this requires a bit of trickery and so the code is a little more
# convoluted than it might otherwise be.
class ConcreteExprEditor(AggregatedEditor):

    # return an abstract expr (a string of the form "X(y1,y2,...,yn)"
    # corresponding to the current concrete expr that is the value of the
    # thing we are editing
    def getAbsExpr(self):
        raise NotImplementedError("ConcreteExprEditor.getAbsExpr() is abstract!")

    # return a tuple containing the current values for the arguments
    # of the thing we're editing
    def getCurrentArgs(self):
        raise NotImplementedError("ConcreteExprEditor.getCurrentArgs() is abstract")

    def updateImpl(*args):
        raise NotImplementedError("ConcreteExprEditor.updateImpl() is abstract")
    def addImpl(*args):
        raise NotImplementedError("ConcreteExprEditor.addImpl() is abstract")

    def getDesc(self):
        raise NotImplementedError("ConcreteExprEditor.getDesc() is abstract")
    
    
    # create an editor for 'name'
    #    name is the thing we're editing. if None, then we build an editor that
    #       can be used to specify a new concrete expr; we include an Entry
    #       that allows the user to give the expr a name
    #    systemeditor is the SystemEditor instance that created us
    #    widgetChangedCallback is a callback to be called when our
    #       widget changes, which will occur whenever we switch which
    #       sub-editor is registered (i.e. when the user makes a selection
    #       in the combobox)
    def __init__(self, name, systemeditor, widgetChangedCallback, listStore):
        AggregatedEditor.__init__(self)
        self.systemeditor = systemeditor
        self.name = name
        self.widgetChangedCallback = widgetChangedCallback
        self.listStore = listStore
        
        self.combobox = gtk.ComboBox(listStore)
        cell = gtk.CellRendererText()
        self.combobox.pack_start(cell, True)
        self.combobox.add_attribute(cell, "text", 0)
        self.combobox.connect("changed", self.comboChanged, None)
        # a dictionary of parameter editors -- one for each abs exprs
        # these are created on the fly and reused as needed
        self.paramsEditors = {}
        
        self.currentEditor = None

        if self.name != None:
            # get the abstract expr that corresponds to the actual
            # concrete expr of the thing we're editing
            initialAbsExpr = self.getAbsExpr()


            # set the initial selection of the combobox
            # this results in a call to comboChanged which will create
            # the appropriate editor for the current abstract expr
            try:
                self.setTextInCombobox(self.combobox, initialAbsExpr)
            except Exception:
                # if error, it's probably because the abstract expr
                # has been deleted by the user. in this case, build an
                # editor anyway
                self.comboChanged(self.combobox, None)
        else:
            # we are an editor for adding new concrete exprs
            # include an Entry to allow the user to give the expr a name
            self.nameEntry = gtk.Entry()
            # we have to manually force the initial editor to be created
            # since there is no initial selection for the combobox
            self.comboChanged(self.combobox, None)
            
    # called whenever the selection in the combobox is altered
    def comboChanged(self, combobox, arg):
        assert arg == None and combobox == self.combobox

        # find out what was selected
        text = self.combobox.get_active_text()
        editor = None

        # since the combobox is shared amongst all these editors,
        # we need to remove it from the current one (if it exists)
        # the same applies to self.nameEntry (if self.name == None)
        if self.currentEditor != None:
            self.currentEditor.detatchWidgetForParam(0)
            if self.name == None:
                self.currentEditor.detatchWidgetForParam(1)

        # have we already created this editor
        if text in self.paramsEditors.keys():
            editor = self.paramsEditors[text]
            # reattach the combobox and the nameEntry if applicable
            editor.reattatchWidgetForParam(0)
            if self.name == None:
                editor.reattatchWidgetForParam(1)
        else:
            # make an editor for this abstract expr that's been selected
            editor = self.makeParamsEditor(text)
            self.paramsEditors[text] = editor

        # unregister the current (i.e. old) editor and add the new one
        if self.currentEditor != None:
            self.removeEditor(self.currentEditor)
        self.addEditor(editor)

        
        if self.name != None:
            # get the abs expr corresponding to the thing we're editing
            absExpr = self.getAbsExpr()
            
            # tell this editor what it's initial value should be 
            # so it can figure out whether it's in a modified state or not
            editor.modifyInitialValueOfParam(0, absExpr)
            
        self.currentEditor = editor
        
        # notify that our widget has changed
        self.widgetChangedCallback(self.currentEditor.getWidget())

    # make an InvocationEditor for the given abstract expr
    def makeParamsEditor(self, absExpr):
        # parse the abs expr, make a text field for each of the
        # parameters and use the combobox; whack all these into a single
        # InvocationEditor
        # if self.name is None (i.e. we are an editor for describing a
        # new expr), then we also include self.nameEntry
        # to be used so the user can give the new expr a name
        
        invocEdParams = ()
        if self.name == None:
            # we start with a Entry for the name, if required
            invocEdParams += ( ("Name: ", self.nameEntry, None,
                                getTextFromEntry, setTextInEntry), )
            
        # combobox goes before all parameters
        invocEdParams += ( (None, self.combobox, None,
                           self.getTextFromCombobox,
                           self.setTextInCombobox), )

        # absExpr can be None if this is an editor for no abstract expr
        # (i.e. for when no selection has been made on the combobox)
        if absExpr != None:
            (name, params) = expr.parse(absExpr)
        else:
            params = None

        if self.name != None:
            currArgs = self.getCurrentArgs()
            currAbsExpr = self.getAbsExpr()
        entryValues = None
        if self.name != None and currAbsExpr != None and currAbsExpr == absExpr:
            # the user has (re)selected the current abstract expr
            # so fill in the text fields with the current values
            # to reflect the actual state of things
            entryValues = currArgs
        # add each parameter
        index = 0
        if params != None:
            for param in params:
                pentry = gtk.Entry()
                if entryValues != None:
                    pentry.set_text(entryValues[index])
                invocEdParams += ( ("%s: " % param, pentry, None, getTextFromEntry,
                                    setTextInEntry), )
                index += 1

        desc = None
        if self.name != None:
            op = "_Commit"
            desc = "Update %s" % self.getDesc()
            impl = self.updateImpl
        else:
            op = "_Add" 
            desc = "Add %s" % self.getDesc()
            impl = self.addImpl
        orig = None
        if self.name != None:
            orig = (currAbsExpr,) + ((None,) * index)
        else:
            orig = ("", None) + ((None,) * index)
            
        editor = InvocationEditor(op,
                                  desc,
                                  impl,
                                  invocEdParams, self.name == None,
                                  orig)
        return editor

    def getTextFromCombobox(self, combobox):
        assert combobox == self.combobox
        assert combobox.get_model() == self.listStore
        return combobox.get_active_text()

    def setTextInCombobox(self, combobox, text):
        assert combobox == self.combobox
        assert combobox.get_model() == self.listStore
        store = self.listStore
        iter = store.get_iter_root()
        while iter != None:
            name = store.get(iter, 0)[0]
            if name == text:
                break
            iter = store.iter_next(iter)
        else:
            raise Exception("Internal Error: %s not found in combobox" % text)
        combobox.set_active_iter(iter)

    def getWidget(self):
        assert self.currentEditor != None
        return self.currentEditor.getWidget()


# editor for a specific object's behaviour.
class ObjectBehaviourEditor(ConcreteExprEditor):

    # get the abstract behaviour for the object we're editing, if none
    # exists (which can happen if the object's abstract behaviour has
    # been removed), returns None
    def getAbsExpr(self):
        behaviour = self.systemeditor.system.objects[self.name]
        try:
            (absBehaviour, map) = behaviour.findAbsBehaviour(self.systemeditor.system)
            return absBehaviour
        except ValueError:
            return None

    def getCurrentArgs(self):
        return self.systemeditor.system.objects[self.name].args
    
    # create an editor for an object's behaviour
    #    name is the object's name. if None, then we build an editor that
    #       can be used to specify the behaviour of a new object in the
    #       system; we include an Entry that allows the user to give the
    #       new object a name
    #    systemeditor is the SystemEditor instance that created us
    #    widgetChangedCallback is a callback to be called when our
    #       widget changes, which will occur whenever we switch which
    #       sub-editor is registered (i.e. when the user makes a selection
    #       in the combobox)
    def __init__(self, name, systemeditor, widgetChangedCallback):
        ConcreteExprEditor.__init__(self, name, systemeditor, widgetChangedCallback, systemeditor.absBehavioursListStore)
            

    def getDesc(self):
        if self.name == None:
            return "An Object"
        else:
            return "%s's Behaviour" % self.name
        
    # called to update the behaviour of the object (self.name != None)
    def updateImpl(*args):

        self = args[0]
        assert self.name != None
        name = args[1]
        args = args[2:]
        (name, absArgs) = expr.parse(name)
        if args == ():
            args = None
        behaviour = Behaviour(name,args)
        return self.systemeditor.performAndLog(self.systemeditor.system.updateObject,
                                               (self.name, behaviour),
                                               "Updating behaviour of object %s" % self.name)
        
        
    # called to add a new object to the system (self.name == None)
    def addImpl(*args):
        def do(*args):
            self = args[0]
            assert self.name == None
            objname = args[1]

            absBehaviourName = args[2]
            if absBehaviourName == None:
                raise ValueError("Cannot use empty name for object behaviour")
            
            args = args[3:]
            (absBehaviourName, absArgs) = expr.parse(absBehaviourName)
            if args == ():
                args = None
            behaviour = Behaviour(absBehaviourName,args)
            if self.systemeditor.performAndLog(self.systemeditor.system.addObject,
                                               (objname, behaviour),
                                               "Adding object %s" % objname):
                
                # update the system tree
                self.systemeditor.addNodeToSystemTreeView(
                    util.seqString((SYSTEM_NAME,
                                    NODE_OBJECTS,
                                    objname),FQ_NAME_SEP))
                return True
            return False
                    
        self = args[0]
        assert self.name == None
        
        return self.systemeditor.performAndLog(do,args,
                                               "Adding object")

# editor for the refines of an absbehaviour.
class AbsBehaviourRefinesEditor(ConcreteExprEditor):

    # get the abstract behaviour that corresponds to the specified refines
    # that we're editing -- if it doesn't 
    # exists (which can happen if the abstract behaviour has
    # been removed), returns None
    def getAbsExpr(self):
        try:
            behaviour = self.systemeditor.system.refines[self.name]
        except KeyError:
            return None
        try:
            (absBehaviour, map) = behaviour.findAbsBehaviour(self.systemeditor.system)
            return absBehaviour
        except ValueError:
            return None

    def getCurrentArgs(self):
        try:
            return self.systemeditor.system.refines[self.name].args
        except KeyError:
            return None
    
    # create an editor for the refines for an abstract behaviour
    #    name is the absBehaviour name.
    #    systemeditor is the SystemEditor instance that created us
    #    widgetChangedCallback is a callback to be called when our
    #       widget changes, which will occur whenever we switch which
    #       sub-editor is registered (i.e. when the user makes a selection
    #       in the combobox)
    def __init__(self, name, systemeditor, widgetChangedCallback):
        assert name != None
        ConcreteExprEditor.__init__(self, name, systemeditor, widgetChangedCallback, systemeditor.absBehavioursListStoreWithNone)
            

    def getDesc(self):
        return "refines statement for %s" % self.name
        
    # called to update the behaviour of the object (self.name != None)
    def updateImpl(*args):
        self = args[0]
        assert self.name != None
        name = args[1]
        behaviour = None
        desc = "Removing refines statement for %s" % self.name
        if name != None:
            args = args[2:]
            (name, absArgs) = expr.parse(name)
            if args == ():
                args = None
            behaviour = Behaviour(name,args)
            desc = "Updating refines statement for %s" % self.name
        return self.systemeditor.performAndLog(self.systemeditor.system.updateRefines,
                                               (self.name, behaviour),
                                               desc)


# editor for a specific assertion
class AssertionEditor(ConcreteExprEditor):

    def getAbsExpr(self):
        assertion = self.systemeditor.system.assertions[self.name]
        try:
            (absPred, map) = self.systemeditor.predicates.findPredicate(assertion)
            return absPred
        except ValueError:
            return None

    def getCurrentArgs(self):
        return self.systemeditor.system.assertions[self.name][1]
    
    def __init__(self, name, systemeditor, widgetChangedCallback):
        ConcreteExprEditor.__init__(self, name, systemeditor, widgetChangedCallback, systemeditor.predicatesListStore)
            
    def getDesc(self):
        if self.name == None:
            return "An Assertion"
        else:
            return self.name

    # called to update the assertion
    def updateImpl(*args):
        self = args[0]
        assert self.name != None
        name = args[1]
        args = args[2:]
        (name, absArgs) = expr.parse(name)
        if args == ():
            args = None
        assertion = (name, args)
        return self.systemeditor.performAndLog(self.systemeditor.system.updateAssertion,
                                               (self.name, assertion),
                                               "Updating assertion %s" % self.name)
    # called to add a new assertion to the system (self.name == None)
    def addImpl(*args):
        def do(*args):
            self = args[0]
            assert self.name == None
            assname = args[1]
            predicateName = args[2]
            args = args[3:]
            (predicateName, absArgs) = expr.parse(predicateName)
            if args == ():
                args = None
            assertion = (predicateName, args)
            if self.systemeditor.performAndLog(self.systemeditor.system.addAssertion,
                                               (assname, assertion),
                                               "Adding assertion %s" % assname):
                
                # update the system tree
                self.systemeditor.addNodeToSystemTreeView(
                    util.seqString((SYSTEM_NAME,
                                    NODE_ASSERTIONS,
                                    assname),FQ_NAME_SEP))
                return True
            return False
                    
        self = args[0]
        assert self.name == None
        
        return self.systemeditor.performAndLog(do,args,
                                               "Adding assertion")    
    
class SystemAssertionEditor(AggregatedEditor):
    def __init__(self, fqName, systemeditor, name):
        AggregatedEditor.__init__(self)
        self.fqName = fqName
        self.systemeditor = systemeditor
        self.name = name
        (self.frame, self.vbox) = util.makeVBoxInFrame(self.fqName,
                                             "Edit assertion %s." % self.name)

        self.assertionEditorWidget = None
        self.assertionEditorWidgetContainer = gtk.VBox()
        self.vbox.pack_start(self.assertionEditorWidgetContainer, False, False, 0)
        self.assertionEditor = AssertionEditor(name, systemeditor,
                                               self.assertionEditorWidgetChanged)

        self.addEditor(self.assertionEditor)


        checkMe = Action("Chec_k", "Check Assertion", self.checkMe,
                           util.STOCK_CHECK)
        self.vbox.pack_start(checkMe.getWidget(), False, False, 0)
        self.addModifiedCallback(lambda ed : checkMe.setSensitive(False))
        self.addSyncedCallback(lambda ed : checkMe.setSensitive(True))

        delMe = Action(None, "Delete %s" % name, self.deleteMe,
                       gtk.STOCK_DELETE)
        self.vbox.pack_start(delMe.getWidget(), False, False, 0)

    def checkMe(self):
        self.systemeditor.checkAssertion(self.name)
        
    def assertionEditorWidgetChanged(self, newWidget):
        if self.assertionEditorWidget != None:
            self.assertionEditorWidgetContainer.remove(self.assertionEditorWidget)
        self.assertionEditorWidgetContainer.pack_start(newWidget, False, False, 0)
        self.systemeditor.redrawWindow()
        self.assertionEditorWidget = newWidget
        
    def getWidget(self):
        return self.frame

    # callback for when this assertion is to be deleted
    def deleteMe(self):
        if self.systemeditor.performAndLog(self.systemeditor.system.removeAssertion, (self.name,),
                                           "Removing assertion %s" % self.name):
            self.systemeditor.removeNodeFromSystemTreeView(self.fqName)
            self.systemeditor.removeEditor(self)

    
        
# editor for the authority that an object is trusted with
class AuthorityEditor(InvocationEditor):

    def __init__(self, name, initialAuthority, commitCallback, redrawWindow):
        self.redrawWindow = redrawWindow
        
        # we use a number of gtk.Entry fields to specify the object's
        # authority of the various kinds
        self.permsEntry = gtk.Entry()   # acquirable permissions
        self.spcEntry = gtk.Entry()     # strong positive causation
        self.sncEntry = gtk.Entry()     # strong negative causation
        self.wpcEntry = gtk.Entry()     # weak positive causation
        self.wncEntry = gtk.Entry()     # weak negative causation

        # ignore direct causation for now
        #self.dpcEntry = gtk.Entry()     # direct positive causation
        #self.dncEntry = gtk.Entry()     # direct negative causation

        # an object's strong (positive|negative) causation authority
        # should be a subset of its weak (positive|negative) authority,
        # since it should never be the case that an object should be
        # allowed to strongly cause an event to occur but not weakly cause
        # it to occur

        self.useStrongCausationCheck = gtk.CheckButton("_Include _Strong Causation")
            
        self.useNegCausationCheck = gtk.CheckButton("Include _Negative Causation")
        # ignore direct causation for now
        #self.useDirCausationCheck = gtk.CheckButton("Include _Direct Causation")        

        if initialAuthority != None:
            if initialAuthority.perms != None:
                self.permsEntry.set_text(initialAuthority.perms)
            if initialAuthority.wpc != None:
                self.wpcEntry.set_text(initialAuthority.wpc)
            if initialAuthority.wnc != None:
                self.wncEntry.set_text(initialAuthority.wnc)
            if initialAuthority.spc != None:
                self.spcEntry.set_text(initialAuthority.spc)
            if initialAuthority.snc != None:
                self.sncEntry.set_text(initialAuthority.snc)
            #if initialAuthority.dpc != None:
            #    self.dpcEntry.set_text(initialAuthority.dpc)
            #if initialAuthority.dnc != None:
            #    self.dncEntry.set_text(initialAuthority.dnc)
                
            if initialAuthority.snc != None or initialAuthority.spc != None:
                self.useStrongCausationCheck.set_active(True)
            
            if initialAuthority.wnc != None or initialAuthority.snc != None:
                self.useNegCausationCheck.set_active(True)

            #if initialAuthority.dnc != None or initialAuthority.dpc != None:
            #    self.useDirCausationCheck.set_active(True)
                
        
        InvocationEditor.__init__(self, "_Commit",
                                  "Edit %s's Allowed Authority" % name,
                                  commitCallback,
                                  ((None, self.useStrongCausationCheck,
                                    (self.useStrongCausationCheck, "toggled"),
                                    getBoolFromCheck, setBoolInCheck),
                                   (None, self.useNegCausationCheck,
                                    (self.useNegCausationCheck, "toggled"),
                                    getBoolFromCheck, setBoolInCheck),
                                   #(None, self.useDirCausationCheck,
                                   # (self.useDirCausationCheck, "toggled"),
                                   # getBoolFromCheck, setBoolInCheck),
                                   ("Acquirable Permissions: ",
                                    self.permsEntry, None,
                                    getTextFromEntry, setTextInEntry),
                                   ("Weak Positive Causation: ",
                                    self.wpcEntry, None,
                                    getTextFromEntry, setTextInEntry),
                                   ("Weak Negative Causation: ",
                                    self.wncEntry, None,
                                    getTextFromEntry, setTextInEntry),
                                   ("Strong Positive Causation: ",
                                    self.spcEntry, None,
                                    getTextFromEntry, setTextInEntry),
                                   ("Strong Negative Causation: ",
                                    self.sncEntry, None,
                                    getTextFromEntry, setTextInEntry)
                                   #,
                                   # ignore direct causation for now
                                   #("Direct Positive Causation: ",
                                   # self.dpcEntry, None,
                                   # getTextFromEntry, setTextInEntry),
                                   #("Direct Negative Causation: ",
                                   # self.dncEntry, None,
                                   # getTextFromEntry, setTextInEntry)
                                   ))

        # this is a bit of a hack here -- we build the editor initially
        # with all Entrys, so this corresponds to the user wanting to
        # use strong, negative and direct causation, hence we set these True
        self.useStrongCausation = True
        self.useNegCausation = True

        # since we're ignoring d[pn]c, we should ensure this stays false
        self.useDirCausation = False

        # we then set the checkbuttons to match, which may
        # cause the editor to go into a modified state
        self.useStrongCausationCheck.set_active(self.useStrongCausation)
        self.useNegCausationCheck.set_active(self.useNegCausation)
        #self.useDirCausationCheck.set_active(self.useDirCausation)        

        # we then connect our own callbacks to the checkbuttons so we can
        # update our appearance (add and remove widgets) as the user changes
        # their preference about the kinds of authority they're interested in
        self.useStrongCausationCheck.connect("toggled", self.useStrongCausationToggled)
        self.useNegCausationCheck.connect("toggled", self.useNegCausationToggled)
        #self.useDirCausationCheck.connect("toggled", self.useDirCausationToggled)        
        # since we're ignoring direct causation, these widgets come earlier
        self.wpcIndex = 4 - 1
        self.wncIndex = 5 - 1
        self.spcIndex = 6 - 1
        self.sncIndex = 7 - 1
        #self.dpcIndex = 8
        #self.dncIndex = 9

        # now set the buttons to their appropriate initial values -- this will
        # put the editor
        # back into the synced state which is what we want. it will also
        # ensure the editor starts out in the default, minimal state
        # since our callbacks will be called to remvoe the unnecessary
        # widgets
        useStrong = initialAuthority != None and (initialAuthority.snc != None or initialAuthority.spc != None)
        if not useStrong:
            self.useStrongCausationCheck.set_active(False)
            
        useNeg = initialAuthority != None and (initialAuthority.snc != None or initialAuthority.wnc != None)
        if not useNeg:
            self.useNegCausationCheck.set_active(False)
            
        #useDir = initialAuthority != None and (initialAuthority.dnc != None or initialAuthority.dpc != None)
        #if not useDir:
        #    self.useDirCausationCheck.set_active(False)
        
        

    def useStrongCausationToggled(self, button):
        assert button == self.useStrongCausationCheck

        use = button.get_active()
        assert use == (not self.useStrongCausation)
        if not use:
            self.detatchLabelForParam(self.spcIndex)
            self.detatchWidgetForParam(self.spcIndex)
            if self.useNegCausation:
                self.detatchLabelForParam(self.sncIndex)
                self.detatchWidgetForParam(self.sncIndex)
        else:
            self.reattatchLabelForParam(self.spcIndex)
            self.reattatchWidgetForParam(self.spcIndex)
            if self.useNegCausation:
                self.reattatchLabelForParam(self.sncIndex)
                self.reattatchWidgetForParam(self.sncIndex)
            
        self.useStrongCausation = use
        self.redrawWindow()
        
    def useNegCausationToggled(self, button):
        assert button == self.useNegCausationCheck

        use = button.get_active()
        assert use == (not self.useNegCausation)
        if not use:
            self.detatchLabelForParam(self.wncIndex)
            self.detatchWidgetForParam(self.wncIndex)
            if self.useStrongCausation:
                self.detatchLabelForParam(self.sncIndex)
                self.detatchWidgetForParam(self.sncIndex)
            if self.useDirCausation:
                self.detatchLabelForParam(self.dncIndex)
                self.detatchWidgetForParam(self.dncIndex)
        else:
            self.reattatchLabelForParam(self.wncIndex)
            self.reattatchWidgetForParam(self.wncIndex)
            if self.useStrongCausation:
                self.reattatchLabelForParam(self.sncIndex)
                self.reattatchWidgetForParam(self.sncIndex)
            if self.useDirCausation:
                self.reattatchLabelForParam(self.dncIndex)
                self.reattatchWidgetForParam(self.dncIndex)

        self.useNegCausation = use
        self.redrawWindow()    

#     def useDirCausationToggled(self, button):
#         assert button == self.useDirCausationCheck

#         use = button.get_active()
#         assert use == (not self.useDirCausation)
#         if not use:
#             self.detatchLabelForParam(self.dpcIndex)
#             self.detatchWidgetForParam(self.dpcIndex)
            
#             if self.useNegCausation:
#                 self.detatchLabelForParam(self.dncIndex)
#                 self.detatchWidgetForParam(self.dncIndex)
#         else:
#             self.reattatchLabelForParam(self.dpcIndex)
#             self.reattatchWidgetForParam(self.dpcIndex)
#             if self.useNegCausation:
#                 self.reattatchLabelForParam(self.dncIndex)
#                 self.reattatchWidgetForParam(self.dncIndex)

#         self.useDirCausation = use
#         self.redrawWindow()    
        
# editor for a specific object
class ObjectEditor(AggregatedEditor):
    def __init__(self, fqName, systemeditor, name):
        AggregatedEditor.__init__(self)
        self.fqName = fqName
        self.systemeditor = systemeditor
        self.name = name
        (self.frame, self.vbox) = util.makeVBoxInFrame(self.fqName,
                                             "Edit object %s." % self.name)

        exploreMe = Action("_Explore", "Explore %s" % name, self.exploreMe,
                           util.STOCK_EXPLORE)
        self.vbox.pack_start(exploreMe.getWidget(), False, False, 0)
        self.addModifiedCallback(lambda ed : exploreMe.setSensitive(False))
        self.addSyncedCallback(lambda ed : exploreMe.setSensitive(True))
        self.behaviourEditorWidget = None
        self.behaviourEditorWidgetContainer = gtk.VBox()
        self.vbox.pack_start(self.behaviourEditorWidgetContainer, False, False, 0)
        self.behaviourEditor = ObjectBehaviourEditor(name, systemeditor,
                                                     self.behaviourEditorWidgetChanged)

        self.addEditor(self.behaviourEditor)


        initialAuthority = None
        if self.name in self.systemeditor.system.allowedAuthority.keys():
            initialAuthority = self.systemeditor.system.allowedAuthority[self.name]
        self.authEdit = AuthorityEditor(self.name, initialAuthority,
                                        self.updateAuthorityImpl,
                                        self.systemeditor.redrawWindow)
        self.vbox.pack_start(self.authEdit.getWidget(), False, False, 0)
        self.addEditor(self.authEdit)

        delMe = Action(None, "Delete %s" % name, self.deleteMe,
                       gtk.STOCK_DELETE)
        self.vbox.pack_start(delMe.getWidget(), False, False, 0)

        
        
    def updateAuthorityImpl(self, useStrong, useNeg,
                            #useDir,
                            perms, wpc, wnc,
                            spc, snc,
                            #dpc, dnc
                            ):
        if not useStrong:
            spc = ""
            snc = ""
        if not useNeg:
            wnc = ""
            snc = ""
#         if not useDir:
#             dpc = ""
#             dnc = ""

        noneCount = 0
        if perms.lstrip() == "":
            perms = None
            noneCount += 1
        if wpc.lstrip() == "":
            wpc = None
            noneCount += 1            
        if wnc.lstrip() == "":
            wnc = None
            noneCount += 1            
        if spc.lstrip() == "":
            spc = None
            noneCount += 1            
        if snc.lstrip() == "":
            snc = None
            noneCount += 1            
#         if dpc.lstrip() == "":
#             dpc = None
#             noneCount += 1            
#         if dnc.lstrip() == "":
#             dnc = None
#             noneCount += 1            
        dpc = None
        dnc = None
        
        if noneCount == 5:
            authority = None
        else:
            assert noneCount < 7
            authority = Authority(perms,wpc,wnc,spc,snc,dpc,dnc)

        
        return self.systemeditor.performAndLog(self.systemeditor.system.updateAllowedAuthority, (self.name, authority),
                                           "Updating allowed authority for object %s" % self.name)
        
        
    def behaviourEditorWidgetChanged(self, newWidget):
        if self.behaviourEditorWidget != None:
            self.behaviourEditorWidgetContainer.remove(self.behaviourEditorWidget)
        self.behaviourEditorWidgetContainer.pack_start(newWidget, False, False, 0)
        self.systemeditor.redrawWindow()
        self.behaviourEditorWidget = newWidget
        
    def getWidget(self):
        return self.frame

    # callback for when this object is to be deleted
    def deleteMe(self):
        if self.systemeditor.performAndLog(self.systemeditor.system.removeObject, (self.name,),
                                           "Removing object %s" % self.name):
            self.systemeditor.removeNodeFromSystemTreeView(self.fqName)
            self.systemeditor.removeEditor(self)

    # callback for when this object is to be explored
    def exploreMe(self):
        self.systemeditor.exploreObject(self.name)
    

# editor for the system's objects
class ObjectsEditor(AggregatedEditor):
    def __init__(self, fqName, systemeditor):
        AggregatedEditor.__init__(self)
        self.fqName = fqName
        self.systemeditor = systemeditor
        
        (frame, vbox) = util.makeVBoxInFrame(self.fqName,
                                        getEditString(NODE_OBJECTS))
        self.vbox = vbox

        self.behaviourEditorWidget = None        
        self.addObj = ObjectBehaviourEditor(None, self.systemeditor,
                                       self.behaviourEditorWidgetChanged)
        
        self.addEditor(self.addObj)
        self.widget = frame
    def getWidget(self):
        return self.widget
        
    def behaviourEditorWidgetChanged(self, newWidget):
        if self.behaviourEditorWidget != None:
            self.vbox.remove(self.behaviourEditorWidget)
        self.vbox.pack_start(newWidget, False, False, 0)
        self.systemeditor.redrawWindow()
        self.behaviourEditorWidget = newWidget
        

# editor for the system's assertions
class AssertionsEditor(AggregatedEditor):
    def __init__(self, fqName, systemeditor):
        AggregatedEditor.__init__(self)
        self.fqName = fqName
        self.systemeditor = systemeditor
        
        (frame, vbox) = util.makeVBoxInFrame(self.fqName,
                                        getEditString(NODE_ASSERTIONS))
        self.vbox = vbox

        self.assertionEditorWidget = None        
        self.addAssertion = AssertionEditor(None, self.systemeditor,
                                            self.assertionEditorWidgetChanged)
        
        self.addEditor(self.addAssertion)
        self.widget = frame

    def getWidget(self):
        return self.widget
        
    def assertionEditorWidgetChanged(self, newWidget):
        if self.assertionEditorWidget != None:
            self.vbox.remove(self.assertionEditorWidget)
        self.vbox.pack_start(newWidget, False, False, 0)
        self.systemeditor.redrawWindow()
        self.assertionEditorWidget = newWidget
        
        

# editor for those parts of the system that aren't editable!    
class NotEditableEditor(AggregatedEditor):
    def __init__(self, fqName, systemeditor):
        Editor.__init__(self)
        self.fqName = fqName
        self.systemeditor = systemeditor
        
        (frame, vbox) = util.makeVBoxInFrame(fqName,
                                        getNotEditableString(fqName))
        
        self.widget = frame
        
    def getWidget(self):
        return self.widget
    
    
class SystemEditor(AggregatedEditor):

    # called to remove a sub-editor
    def removeEditor(self, editor):
        assert editor in self.inverseEditorsMap
        fqName = self.inverseEditorsMap[editor]
        assert fqName in self.editorsMap and self.editorsMap[fqName] == editor
        del self.editorsMap[fqName]
        del self.inverseEditorsMap[editor]
        # we may not have registered this editor with
        # AggregatedEditor.addEditor()
        if editor in self.editors:
            AggregatedEditor.removeEditor(self, editor)
        
    # override the method to handle when a subeditor is modified
    # so we can capture this information
    def subEditorModified(self, editor):
        node = self.inverseEditorsMap[editor]
        self.setSystemTreeNodeStatus(node, STATUS_MODIFIED)
        # propoagate the notification
        AggregatedEditor.subEditorModified(self, editor)

    def subEditorSynced(self, editor):
        node = self.inverseEditorsMap[editor]
        self.setSystemTreeNodeStatus(node, STATUS_SYNCED)        
        # propagate the notification
        AggregatedEditor.subEditorSynced(self, editor)
        


    # make the ListStore that represents the different abstract predicates
    def makePredicatesListStore(self):
        assert self.predicatesListStore == None
        store = gtk.ListStore(gobject.TYPE_STRING)

        for pred in sorted(self.predicates.getPredicateNames()):
            store.append((pred,))
        self.predicatesListStore = store

    # make the ListStore that represents the different abstract
    # behaviour names for our system
    def makeAbsBehavioursListStore(self):
        assert self.absBehavioursListStore == None
        assert self.absBehavioursListStoreWithNone == None
        
        store = gtk.ListStore(gobject.TYPE_STRING)
        storeWithNone = gtk.ListStore(gobject.TYPE_STRING)
        storeWithNone.append((None,))        
        for absBehaviour in sorted(self.system.behaviours.keys()):
            store.append((absBehaviour,))
            storeWithNone.append((absBehaviour,))            
        self.absBehavioursListStore = store
        self.absBehavioursListStoreWithNone = storeWithNone

    # add an entry to the abs behaviours list store
    def addNodeToAbsBehavioursListStore(self, absBehaviour):
        def addNameToListStore(theName, store):
            iter = store.get_iter_root()
            if iter == None:
                return store.append((theName,))
            else:
                while iter != None:
                    name = store.get(iter, 0)[0]
                    res = cmp(name, theName)
                    if res < 0:
                        ()
                    elif res > 0:
                        # found it
                        return store.insert_before(iter,(theName,))
                    else:
                        raise Exception("Internal Error: cannot add dupliate name %s" % name)
                    iter = store.iter_next(iter)
                # we go at the end
                return store.append((theName,))
                    
                
        for store in (self.absBehavioursListStore,
                      self.absBehavioursListStoreWithNone):
            addNameToListStore(absBehaviour, store)

    # remove an entry from the abs behaviours list store
    def removeNodeFromAbsBehavioursListStore(self, absBehaviour):
        def removeNameFromListStore(name, store):
            iter = store.get_iter_root()
            while iter != None:
                name = store.get(iter, 0)[0]
                if name == absBehaviour:
                    store.remove(iter)
                    break
                iter = store.iter_next(iter)
            else:
                raise Exception("Internal Error: cannot remove nonexistent node %s" % node)
        for store in (self.absBehavioursListStore,
                      self.absBehavioursListStoreWithNone):
            removeNameFromListStore(absBehaviour, store)
            
        
        
    # make the TreeView that represents the given system
    def makeSystemTreeView(self):
        assert self.systemTreeView == None
        store = gtk.TreeStore(gobject.TYPE_STRING, gobject.TYPE_STRING)
        treeview = gtk.TreeView(store)

        # have the name of the system as the root node
        iter = store.append(None,  (SYSTEM_NAME, STATUS_SYNCED))
        root = iter
        
        # add the operations
        iter = store.append(root, (NODE_OPS, STATUS_SYNCED))
        for op in sorted(self.system.ops):
            child = store.append(iter, (op, STATUS_SYNCED))

        # add the objects
        iter = store.append(root, (NODE_OBJECTS, STATUS_SYNCED))
        for obj in sorted(self.system.objects.keys()):
            child = store.append(iter, (obj, STATUS_SYNCED))

        # add the behaviours
        iter = store.append(root, (NODE_ABSBEHAVIOURS, STATUS_SYNCED))
        for absBehaviour in sorted(self.system.behaviours.keys()):
            child = store.append(iter, (absBehaviour, STATUS_SYNCED))


        # add the assertions
        iter = store.append(root, (NODE_ASSERTIONS, STATUS_SYNCED))
        for assertion in sorted(self.system.assertions.keys()):
            child = store.append(iter, (assertion, STATUS_SYNCED))

        # add the code
        iter = store.append(root, (NODE_CODE, STATUS_SYNCED))

        # add the columns to the treeview
        cell = gtk.CellRendererText()
        column = gtk.TreeViewColumn("Component", cell, text=0)
        treeview.append_column(column)
        italicCell = gtk.CellRendererText()
        italicCell.set_property("style", pango.STYLE_ITALIC)
        column = gtk.TreeViewColumn("Status", italicCell, text=1)
        treeview.append_column(column)

        selection = treeview.get_selection()
        selection.connect("changed", self.systemTreeViewSelectChanged)

        treeview.expand_all()
            
        self.systemTreeView = treeview
        self.systemTreeStore = store
        self.systemTreeViewContainer.add(self.systemTreeView)

    # given a string representing a fully qualified path name in the
    # system tree view, add the nodes required to put tihis path in the tree
    def addNodeToSystemTreeView(self, fqName):
        def addNameToTreeStore(theName, parent, store):
            iter = store.iter_children(parent)
            if iter == None:
                return store.append(parent, (theName, STATUS_SYNCED))
            else:
                while iter != None:
                    name = store.get(iter, 0)[0]
                    res = cmp(name, theName)
                    if res < 0:
                        ()
                    elif res > 0:
                        # found it
                        return store.insert_before(parent, iter, (theName,
                                                           STATUS_SYNCED))
                    else:
                        raise Exception("Internal Error: cannot add dupliate name %s" % name)
                    iter = store.iter_next(iter)
                # we need to go at the end of the list
                return store.append(parent, (theName, STATUS_SYNCED))

        
        
        model = self.systemTreeStore
        names = fqName.split(FQ_NAME_SEP)
        parent = None
        iter = model.get_iter_root()
        for name in names:
            if iter == None:
                # add the node since it doesn't exist
                iter = addNameToTreeStore(name, parent, model)
            else:
                while iter != None and name != model.get(iter, 0)[0]:
                    iter = model.iter_next(iter)
                if iter == None:
                    # add the node since it doesn't exist
                    iter = addNameToTreeStore(name, parent, model)
            parent = iter
            iter = model.iter_children(iter)
                   
    def setSystemTreeNodeStatus(self, node, status):
        model = self.systemTreeStore
        names = node.split(FQ_NAME_SEP)
        parent = None
        iter = model.get_iter_root()
        for name in names:
            if iter == None:
                raise Exception("Internal Error: cannot set status of nonexistent node %s" % node)
            else:
                while iter != None and name != model.get(iter, 0)[0]:
                    iter = model.iter_next(iter)
                if iter == None:
                    raise Exception("Internal Error: cannot set status of nonexistent node %s" % node)                
            parent = iter
            iter = model.iter_children(iter)
        model.set(parent,1,status)

    def removeNodeFromSystemTreeView(self, node):
        model = self.systemTreeStore
        names = node.split(FQ_NAME_SEP)
        parent = None
        iter = model.get_iter_root()
        for name in names:
            if iter == None:
                raise Exception("Internal Error: cannot remove nonexistent node %s" % node)
            else:
                while iter != None and name != model.get(iter, 0)[0]:
                    iter = model.iter_next(iter)
                if iter == None:
                    raise Exception("Internal Error: cannot remove nonexistent node %s" % node)                
            parent = iter
            iter = model.iter_children(iter)

        model.remove(parent)
        
    # update the right-hand pane
    def updateRightPane(self):
        editor = None
        if self.fqSelectedSystemTreeNode != None:
            if self.fqSelectedSystemTreeNode in self.editorsMap.keys():
                editor = self.editorsMap[self.fqSelectedSystemTreeNode]
            else:
                # build the editor
                # which node is selected?
                if self.fqSelectedSystemTreeNode == SYSTEM_NAME:
                    editor = SystemPropertiesEditor(SYSTEM_NAME, self)
                    # notify us when this editor changes                    
                    self.addEditor(editor)
                else:
                    firstPartLen = len(SYSTEM_NAME + FQ_NAME_SEP)
                    qSelectedNode = self.fqSelectedSystemTreeNode[firstPartLen:]

                    if qSelectedNode == NODE_OPS:
                        editor = OperationsEditor(self.fqSelectedSystemTreeNode,self)
                    elif qSelectedNode == NODE_ABSBEHAVIOURS:
                        editor = AbsBehavioursEditor(self.fqSelectedSystemTreeNode,self)
                    elif qSelectedNode == NODE_OBJECTS:
                        editor = ObjectsEditor(self.fqSelectedSystemTreeNode,self)
                    elif qSelectedNode ==  NODE_CODE:
                        editor = CodeEditor(self.fqSelectedSystemTreeNode,self)
                        # notify us when this editor changes
                        self.addEditor(editor)
                    elif qSelectedNode ==  NODE_ASSERTIONS:
                        editor = AssertionsEditor(self.fqSelectedSystemTreeNode,self)
                    elif qSelectedNode.startswith(NODE_OPS + FQ_NAME_SEP):
                        editor = OperationEditor(self.fqSelectedSystemTreeNode,
                                                 self,
                                                 self.selectedSystemTreeNode)
                    elif qSelectedNode.startswith(NODE_ABSBEHAVIOURS + FQ_NAME_SEP):
                        editor = AbsBehaviourEditor(self.fqSelectedSystemTreeNode,
                                                    self,
                                                    self.selectedSystemTreeNode)
                        # notify us when this editor changes
                        self.addEditor(editor)
                    elif qSelectedNode.startswith(NODE_OBJECTS + FQ_NAME_SEP):
                        editor = ObjectEditor(self.fqSelectedSystemTreeNode,
                                              self,
                                              self.selectedSystemTreeNode)
                        # notify us when this editor changes
                        self.addEditor(editor)
                    elif qSelectedNode.startswith(NODE_ASSERTIONS + FQ_NAME_SEP):
                        editor = SystemAssertionEditor(self.fqSelectedSystemTreeNode,
                                              self,
                                              self.selectedSystemTreeNode)
                        # notify us when this editor changes
                        self.addEditor(editor)
                    else:
                        editor = NotEditableEditor(self.fqSelectedSystemTreeNode,
                                                   self)

                # store the editor for later use 
                self.editorsMap[self.fqSelectedSystemTreeNode] = editor
                self.inverseEditorsMap[editor] = self.fqSelectedSystemTreeNode

        if self.rightPaneViewportContents != None:
            self.rightPaneViewport.remove(self.rightPaneViewportContents)

        if editor != None:
            widget = editor.getWidget()
        else:
            widget = self.noneSelectedLabel
            
        self.rightPaneViewport.add(widget)
        self.redrawWindow()
        self.rightPaneViewportContents = widget

    def systemTreeViewSelectChanged(self, selection):
        # find out what was selected
        (model, iter) = selection.get_selected()
        
        if iter != None:
        
            node = model.get(iter, 0)[0]

            # get the fully qualified name of this node
            name = node
            parent = model.iter_parent(iter)
            while parent != None:
                parentName = model.get(parent, 0)[0]
                name = parentName + " / " + name
                iter = parent
                parent = model.iter_parent(iter)
            
            self.fqSelectedSystemTreeNode = name
            self.selectedSystemTreeNode = node
        else:
            self.fqSelectedSystemTreeNode = None
            self.selectedSystemTreeNode = None
            
        self.updateRightPane()


    def getWidget(self):
        return self.hpane
    
    def __init__(self, system, logWindow, redrawWindow, winWidth, winHeight,
                 exploreObject, checkAssertion, predicates):
        AggregatedEditor.__init__(self)
        self.system = system
        self.winWidth = winWidth
        self.winHeight = winHeight
        self.redrawWindow = redrawWindow
        self.logWindow = logWindow
        self.exploreObject = exploreObject
        self.checkAssertion = checkAssertion
        self.predicates = predicates
        
        self.hpane = gtk.HPaned()

        
        self.leftPane = gtk.VBox()
        self.leftPane.set_border_width(util.BORDER_WIDTH)
        self.leftPane.set_spacing(util.VSPACING)

        self.hpane.add1(self.leftPane)
        self.systemTreeViewContainer = gtk.ScrolledWindow()
        self.systemTreeViewContainer.set_policy(gtk.POLICY_AUTOMATIC,
                                                gtk.POLICY_AUTOMATIC)
        self.leftPane.pack_start(self.systemTreeViewContainer, True, True, 0)
        
        self.rightPane = gtk.ScrolledWindow()
        self.rightPane.set_policy(gtk.POLICY_AUTOMATIC,
                                  gtk.POLICY_AUTOMATIC)
        self.rightPaneViewport = gtk.Viewport()
        self.rightPane.set_border_width(util.BORDER_WIDTH)
        self.rightPane.add(self.rightPaneViewport)
        self.hpane.add2(self.rightPane)
        self.rightPaneViewportContents = None

        self.editorsMap = {}
        self.inverseEditorsMap = {}
        self.hpane.set_position(int(winWidth * (0.33)))
        self.systemTreeView = None
        self.makeSystemTreeView()
        self.absBehavioursListStore = None
        self.absBehavioursListStoreWithNone = None        
        self.makeAbsBehavioursListStore()
        self.predicatesListStore = None        
        self.makePredicatesListStore()
        
        # a label to display when none of the treeview nodes is selected
        self.noneSelectedLabel = gtk.Label("Select a system component from the tree on the right to view and edit its properties.")
        self.noneSelectedLabel.set_line_wrap(True)

        # put the appropriate thing in the right-hand pane
        self.systemTreeViewSelectChanged(self.systemTreeView.get_selection())
        self.redrawWindow()
        

    def performAndLog(self, func, args, desc):
        return self.logWindow.performAndLog(func, args, desc)
    
    def commit(self):
        if not self.performAndLog( AggregatedEditor.commit, (self,), "Committing modifications in editor"):
            raise EditorSyncError("Error committing changes in system editor.")
        assert not self.getModified()

    def revert(self):
        if not self.performAndLog( AggregatedEditor.revert, (self,), "Reverting editor"):
            raise EditorSyncError("Error reverting changes in system editor.")
        assert not self.getModified()
        
